import warnings

import numpy
from skimage.transform import (AffineTransform, EuclideanTransform,
                               SimilarityTransform)

from . import defaults, libcoordio
from .exceptions import CoordIOUserWarning, CoordIOError


# low level, functional conversions

def fieldAngle2Cart(xField, yField):
    """Convert (ZEMAX-style) field angles in degrees to a cartesian point on
    the unit sphere

    Zemax defines field angles like this:
    Positive field angles imply positive slope for the ray in that direction,
    and thus refer to **negative** coordinates on distant objects. ZEMAX converts
    x field angles ( αx ) and y field angles ( αy ) to ray direction cosines
    using the following formulas:

    tanαx = l/n
    tanαy = m/n
    l^2 + m^2 + n^2 = 1

    where l, m, and n are the x, y, and z direction cosines.

    Parameters
    -----------
    xField: scalar or 1D array
        zemax style x field
    yField: scalar or 1D array
        zemax style y field

    Returns
    ---------
    result: list
        [x,y,z] coordinates on unit sphere
    """
    # xField, yField = fieldXY

    # invert field degrees, fields represent slope
    # of ray, so a positivly sloped ray is coming from
    # below its respective plane, the resulting vector
    # from this transform should thus point downward, not up
    tanx = numpy.tan(numpy.radians(-1*xField))
    tany = numpy.tan(numpy.radians(-1*yField))

    # calculate direction cosines
    z = numpy.sqrt(1/(tanx**2 + tany**2 + 1))
    x = tanx * z
    y = tany * z
    # if numpy.isnan(n) or numpy.isnan(l) or numpy.isnan(m):
    #     raise RuntimeError("NaN output [%.2f, %.2f, %.2f] for input [%.2f, %.2f]"%(n, l, m, xField, yField))

    return [x,y,z]


def cart2FieldAngle(x, y, z):
    """Convert cartesian point on unit sphere of sky to (ZEMAX-style) field
    angles in degrees.


    Zemax defines field angles like this:
    Positive field angles imply positive slope for the ray in that direction,
    and thus refer to **negative** coordinates on distant objects. ZEMAX converts
    x field angles ( αx ) and y field angles ( αy ) to ray direction cosines
    using the following formulas:

    tanαx = l/n
    tanαy = m/n
    l^2 + m^2 + n^2 = 1

    where l, m, and n are the x, y, and z direction cosines.

    Parameters
    -----------
    x: scalar or 1D array
    y: scalar or 1D array
    z: scalar or 1D array

    Returns
    ---------
    result: list
        [xField, yField] ZEMAX style field coordinates (degrees)
    """
    # if numpy.abs(numpy.linalg.norm(cartXYZ) - 1) > SMALL_NUM:
    #     raise RuntimeError("cartXYZ must be a vector on unit sphere with L2 norm = 1")
    # l, m, n = x, y, z

    # invert field degrees, fields represent slope
    # of ray, so a positivly sloped ray is coming from
    # below the optical axis, the resulting vector
    # from this transform should thus point downward, not up

    xField = -1 * numpy.degrees(numpy.arctan2(x, z))
    yField = -1 * numpy.degrees(numpy.arctan2(y, z))
    # if numpy.isnan(xField) or numpy.isnan(yField):
    #     raise RuntimeError("NaN output [%.2f, %.2f] for input [%.2f, %.2f, %.2f]"%(xField, yField, cartXYZ[0], cartXYZ[1], cartXYZ[2]))
    return [xField, yField]


def cart2Sph(x, y, z):
    """Convert cartesian coordinates to spherical
    coordinates theta, phi in degrees

    phi is polar angle measure from z axis
    theta is azimuthal angle measured from x axis

    Parameters
    -----------
    x: scalar or 1D array
    y: scalar or 1D array
    z: scalar or 1D array

    Returns
    ---------
    result: list
        [theta, phi] degrees
    """

    # if numpy.abs(numpy.linalg.norm(cartXYZ) - 1) > SMALL_NUM:
    #     raise RuntimeError("cartXYZ must be a vector on unit sphere with L2 norm = 1")
    # if cartXYZ[2] < 0:
    #     raise RuntimeError("z direction must be positive for cartesian field coord")
    # x, y, z = cartXYZ
    theta = numpy.degrees(numpy.arctan2(y, x))
    # wrap theta to be between 0 and 360 degrees
    theta = theta % 360

    phi = numpy.degrees(numpy.arccos(z))
    # if numpy.isnan(theta) or numpy.isnan(phi):
    #     raise RuntimeError("NaN output [%.2f, %.2f] from input [%.2f, %.2f, %.2f]"%(theta, phi, cartXYZ[0], cartXYZ[1], cartXYZ[2]))
    return [theta, phi]


def sph2Cart(theta, phi, r=1):
    """Convert spherical coordinates theta, phi in degrees
    to cartesian coordinates on unit sphere.

    phi is polar angle measure from z axis
    theta is azimuthal angle measured from x axis

    Parameters
    -----------
    theta: scalar or 1D array
        degrees, azimuthal angle
    phi: scalar or 1D array
        degrees, polar angle
    r: scalar
        radius of curvature. Default to 1 for unit sphere

    Returns
    ---------
    result: list
        [x,y,z] coordinates on sphere

    """

    theta, phi = numpy.radians(theta), numpy.radians(phi)
    x = r*numpy.cos(theta) * numpy.sin(phi)
    y = r*numpy.sin(theta) * numpy.sin(phi)
    z = r*numpy.cos(phi)

    # if numpy.isnan(x) or numpy.isnan(y) or numpy.isnan(z):
    #     raise RuntimeError("NaN output [%.2f, %.2f, %.2f] for input [%.2f, %.2f]"%(x, y, z, numpy.degrees(theta), numpy.degrees(phi)))

    return [x, y, z]


def fieldToObserved(x, y, z, altCenter, azCenter, pa):
    """Convert xyz unit-spherical field coordinates to Alt/Az

    the inverse of observetToField

    Az=0 is north, Az=90 is east

    +x points along +RA, +y points along +Dec, +z points along boresight
    from telescope toward sky

    Parameters
    ------------
    x: scalar or 1D array
        unit-spherical x field coord (aligned with +RA)
    y: scalar or 1D array
        unit-spherical y field coord (aligned with +Dec)
    z: scalar or 1D array
        unit-spherical z coord
    altCenter: float
        altitude in degrees, field center
    azCenter: float
        azimuth in degrees, field center
    pa: float
        position angle of field center

    Returns
    --------
    alt : float or 1D array
        altitude of coordinate in degrees
    az : float or 1D array

    """
    cosQ = numpy.cos(numpy.radians(-1*pa))
    sinQ = numpy.sin(numpy.radians(-1*pa))
    rotQ = numpy.array([
        [ cosQ, sinQ, 0],
        [-sinQ, cosQ, 0],
        [    0,    0, 1]
    ])

    coords = numpy.array(
        [x, y, z]
    ).T

    coords = rotQ.dot(coords.T).T

    sinPhi = numpy.sin(-1*numpy.radians(90-altCenter))
    cosPhi = numpy.cos(-1*numpy.radians(90-altCenter))
    rotPhi = numpy.array([
        [1,       0,      0],
        [0,  cosPhi, sinPhi],
        [0, -sinPhi, cosPhi]
    ])
    coords = rotPhi.dot(coords.T).T

    sinTheta = numpy.sin(-1*numpy.radians(90-azCenter))
    cosTheta = numpy.cos(-1*numpy.radians(90-azCenter))
    rotTheta = numpy.array([
        [ cosTheta, sinTheta, 0],
        [-sinTheta, cosTheta, 0],
        [        0,        0, 1]
    ])

    coords = rotTheta.dot(coords.T).T

    x, y, z = coords[:, 0], coords[:, 1], coords[:, 2]
    theta, phi = cart2Sph(x, y, z)

    # convert sph theta, phi to az, alt
    az = -1 * theta
    alt = 90 - phi

    az = az % 360  # wrap to 0,360

    return alt, az


def observedToField(alt, az, altCenter, azCenter, pa):
    """Convert Az/Alt coordinates to cartesian coords on the
    unit sphere with the z axis aligned with field center (boresight)

    Az=0 is north, Az=90 is east

    Resulting Cartesian coordinates have the following
    convention +x points along +RA, +y points along +Dec

    Parameters
    ------------
    alt: float or 1D array
        altitude in degrees, positive above horizon, negative below
    az: float or 1D array
        azimuth in degrees az=0 is north az=90 is east
    altCenter: float
        altitude in degrees, field center
    azCenter: float
        azimuth in degrees, field center
    pa: float
        position angle of field center

    Returns
    --------
    theta : float or 1D array
        azimuthal angle of field coordinate (deg)
    phi : float or 1D array
        polar angle of field coordinat (deg off axis)
    """
    # convert from alt/az to spherical
    phis = 90 - alt  # alt
    thetas = -1 * az  # az

    # work in cartesian frame
    coords = sph2Cart(thetas, phis)
    coords = numpy.array(coords).T

    # rotate the xyz coordinate system about z axis
    # such that -y axis is aligned with the azimuthal angle
    # of the field center

    sinTheta = numpy.sin(numpy.radians(90 - azCenter))
    cosTheta = numpy.cos(numpy.radians(90 - azCenter))
    rotTheta = numpy.array([
        [ cosTheta, sinTheta, 0],
        [-sinTheta, cosTheta, 0],
        [        0,        0, 1]
    ])

    coords = rotTheta.dot(coords.T).T


    # rotate the xyz coordinate system about the x axis
    # such that +z points to the field center.

    sinPhi = numpy.sin(numpy.radians(90 - altCenter))
    cosPhi = numpy.cos(numpy.radians(90 - altCenter))
    rotPhi = numpy.array([
        [1,       0,      0],
        [0,  cosPhi, sinPhi],
        [0, -sinPhi, cosPhi]
    ])
    coords = rotPhi.dot(coords.T).T

    # return coords

    # finally rotate about z by the parallactic angle
    # this puts +RA along +X and +DEC along +Y
    cosQ = numpy.cos(numpy.radians(pa))
    sinQ = numpy.sin(numpy.radians(pa))
    rotQ = numpy.array([
        [ cosQ, sinQ, 0],
        [-sinQ, cosQ, 0],
        [    0,    0, 1]
    ])

    coords = rotQ.dot(coords.T).T


    x = coords[:, 0]
    y = coords[:, 1]
    z = coords[:, 2]

    # finally convert back from cartesian to spherical (Field)
    thetaPhi = cart2Sph(x, y, z)
    thetaPhi = numpy.array(thetaPhi).T
    theta, phi = thetaPhi[:,0], thetaPhi[:,1]
    return theta, phi


def fieldToFocal(thetaField, phiField, site, waveCat):
    """Convert spherical field coordinates to the modeled spherical
    position on the focal plane.  Origin is the M1 vertex.

    Raises warnings for coordinates at large field angles.

    Parameters
    -----------
    thetaField: float or numpy.ndarray
        azimuthal angle of field coordinate (deg)
    phiField: float or numpy.ndarray
        polar angle of field coordinate (deg)
    site: str
        "APO" or "LCO"
    waveCat: str
        "Boss", "Apogee", or "GFA"

    Result
    -------
    xFocal: float or numpy.ndarray
        spherical x focal coord mm
    yFocal: float or numpy.ndarray
       spherical y focal coord mm
    zFocal: float or numpy.ndarray
        spherical z focal coord mm
    R: float
        radius of curvature for sphere
    b: float
        z location of center of sphere
    fieldWarn: bool or boolean array
        True if angle off-axis is large enough to be suspicious
    """

    if site == "APO":
        fieldWarn = numpy.array(phiField) > defaults.APO_MAX_FIELD_R
    if site == "LCO":
        fieldWarn = numpy.array(phiField) > defaults.LCO_MAX_FIELD_R

    if hasattr(fieldWarn, "__len__"):
        if True in fieldWarn:
            warnings.warn(
                "Warning! Coordinate far off telescope optical axis " \
                "conversion may be bogus",
                CoordIOUserWarning
            )
    elif fieldWarn is True:
        warnings.warn(
            "Warning! Coordinate far off telescope optical axis, " \
            "conversion may be bogus",
            CoordIOUserWarning
        )

    direction = "focal"
    R, b, c0, c1, c2, c3, c4 = defaults.getFPModelParams(site, direction, waveCat)

    phiFocal = c0 * phiField + c1 * phiField**3 + c2 * phiField**5 \
        + c3 * phiField**7 + c4 * phiField**9
    # thetaField = thetaFocal by definition
    # convert back to xyz coords
    rTheta = numpy.radians(thetaField)
    rPhi = numpy.radians(180 - phiFocal)
    xFocal = R * numpy.cos(rTheta) * numpy.sin(rPhi)
    yFocal = R * numpy.sin(rTheta) * numpy.sin(rPhi)
    zFocal = R * numpy.cos(rPhi) + b

    return xFocal, yFocal, zFocal, R, b, fieldWarn


def focalToField(xFocal, yFocal, zFocal, site, waveCat):
    """Convert xyz focal position to unit-spherical field position.
    xyzFocal do not need to lie on the modeled sphere, and the spherical
    radius of curvature is not needed, only the origin of the sphere
    is needed (b).

    Raises warnings for coordinates at large field angles.

    Parameters
    -----------
    xFocal: float or 1D array
        x focal coord mm
    yFocal: float or 1D array
        y focal coord mm
    zFocal: float or 1D array
        z focal coord mm, +z points along boresight toward sky.
    site: str
        "APO" or "LCO"
    waveCat: str
        "Boss", "Apogee", or "GFA"

    Result
    -------
    thetaField: float or 1D array
        azimuthal field coordinate (deg)
    phiField: float or 1D array
       polar field coordinate (deg)
    fieldWarn: bool or boolean array
        True if angle off-axis is large enough to be suspicious
    """
    direction = "field"
    R, b, c0, c1, c2, c3, c4 = defaults.getFPModelParams(site, direction, waveCat)
    # note by definition thetaField==thetaFocal
    thetaField = numpy.degrees(numpy.arctan2(yFocal, xFocal))

    # generate focal phis (degree off boresight)
    # angle off-axis from optical axis
    rFocal = numpy.sqrt(xFocal**2 + yFocal**2)
    v = numpy.array([rFocal, zFocal]).T
    v[:, 1] = v[:, 1] - b

    # unit vector pointing toward object on focal plane from circle center
    # arc from vertex to off axis
    v = v / numpy.vstack([numpy.linalg.norm(v, axis=1)] * 2).T
    # FP lands at -Z so, Angle from sphere center towards ground
    downwardZaxis = numpy.array([0, -1])
    # phi angle between ray and optical axis measured from sphere center
    phiFocal = numpy.degrees(numpy.arccos(v.dot(downwardZaxis)))
    phiField = c0 * phiFocal + c1 * phiFocal**3 + c2 * phiFocal**5 \
        + c3 * phiFocal**7 + c4 * phiFocal**9

    if site == "APO":
        fieldWarn = numpy.array(phiField) > defaults.APO_MAX_FIELD_R
    if site == "LCO":
        fieldWarn = numpy.array(phiField) > defaults.LCO_MAX_FIELD_R

    if hasattr(fieldWarn, "__len__"):
        if True in fieldWarn:
            warnings.warn(
                "Warning! Coordinate far off telescope optical axis, conversion may be bogus",
                CoordIOUserWarning
            )
    elif fieldWarn is True:
        warnings.warn(
            "Warning! Coordinate far off telescope optical axis, conversion may be bogus",
            CoordIOUserWarning
        )

    return thetaField, phiField, fieldWarn


def focalToWok(
    xFocal, yFocal, zFocal, positionAngle=0,
    xOffset=0, yOffset=0, zOffset=0, tiltX=0, tiltY=0,
    fpScale=1, projectFlat=True
):
    """Convert xyz focal coordinates in mm to xyz wok coordinates in mm.

    The origin of the focal coordinate system is the
    M1 vertex. focal +y points toward North, +x points toward E.
    The origin of the wok coordinate system is the wok vertex.  -x points
    toward the boss slithead.  +z points from telescope to sky.

    Tilt is applied about x axis then y axis.

    Note: currently zFocal irrelevant using a flat wok model.


    Parameters
    -------------
    xFocal: scalar or 1D array
        x position of object on focal plane mm
        (+x aligned with +RA on image)
    yFocal: scalar or 1D array
        y position of object on focal plane mm
        (+y aligned with +Dec on image)
    zFocal: scalar or 1D array
        z position of object on focal plane mm
        (+z aligned boresight and increases from the telescope to the sky)
    positionAngle: scalar
        position angle deg.  Angle measured from (image) North through East to wok +y.
        So position angle of 45 deg, wok +y points NE
    site : str
        either "APO" or "LCO"
    xOffset: scalar or None
        x position (mm) of wok origin (vertex) in focal coords
        calibrated
    yOffset: scalar
        y position (mm) of wok origin (vertex) in focal coords
        calibratated
    zOffset: scalar
        z position (mm) of wok origin (vertex) in focal coords
        calibratated
    tiltX: scalar
        tilt (deg) of wok about focal x axis at PA=0
        calibrated
    tiltY: scalar
        tilt (deg) of wok about focal y axis at PA=0
        calibrated
    fpScale: scalar
        radial scale multiplier
    projectFlat: bool
        if true, assume flat wok model

    Returns
    ---------
    xWok: scalar or 1D array
        x position of object in wok space mm
        (+x aligned with +RA on image)
    yWok: scalar or 1D array
        y position of object in wok space mm
        (+y aligned with +Dec on image)
    zWok: scalar or 1D array
        z position of object in wok space mm
        (+z aligned boresight and increases from the telescope to the sky)
    """

    # apply calibrated tilts and translation (at PA=0)
    # where they should be fit
    # tilts are defined https://mathworld.wolfram.com/RotationMatrix.html
    # as coordinate system rotations counter clockwise when looking
    # down the positive axis toward the origin
    coords = numpy.array([xFocal, yFocal, zFocal])
    transXYZ = numpy.array([xOffset, yOffset, zOffset])

    rotX = numpy.radians(tiltX)
    rotY = numpy.radians(tiltY)
    # rotation about z axis is position angle
    # position angle is clockwise positive for rotation measured from
    # north to wok +y (when looking from above the wok)
    # however rotation matrices are positinve for counter-clockwise rotation
    # hence the sign flip that's coming
    rotZ = -1*numpy.radians(positionAngle)

    rotMatX = numpy.array([
        [1, 0, 0],
        [0, numpy.cos(rotX), numpy.sin(rotX)],
        [0, -1*numpy.sin(rotX), numpy.cos(rotX)]
    ])

    rotMatY = numpy.array([
        [numpy.cos(rotY), 0, -1*numpy.sin(rotY)],
        [0, 1, 0],
        [numpy.sin(rotY), 0, numpy.cos(rotY)]
    ])

    # rotates coordinate system
    rotMatZ = numpy.array([
        [numpy.cos(rotZ), numpy.sin(rotZ), 0],
        [-numpy.sin(rotZ), numpy.cos(rotZ), 0],
        [0, 0, 1]
    ])


    # first apply rotation about x axis (x tilt)
    coords = rotMatX.dot(coords)
    # next apply rotation about y axis (y tilt)
    coords = rotMatY.dot(coords)
    # apply rotation about z axis (angle of observation)
    coords = rotMatZ.dot(coords)

    # apply translation
    if hasattr(xFocal, "__len__"):
        # list of coords fed in
        transXYZ = numpy.array([transXYZ]*len(xFocal)).T
        xWok, yWok, zWok = coords - transXYZ
    else:
        # single set of xyz coords fed in
        xWok, yWok, zWok = coords - transXYZ

    # scale xyWok from dither analysis results
    xWok = xWok / fpScale # from gfa: 0.9998899
    yWok = yWok / fpScale # from gfa: 0.9998899

    if projectFlat:
        # force all z coords to be positioner height
        # necessary for a flat wok model
        # and this should be removed if we go
        # to curved wok
        # !!!!!!!!!!!!!!!!!
        if hasattr(xFocal, "__len__"):
            zWok = numpy.array([defaults.POSITIONER_HEIGHT]*len(xFocal)).T
        else:
            zWok = defaults.POSITIONER_HEIGHT

    return xWok, yWok, zWok


def wokToFocal(
    xWok, yWok, zWok, positionAngle=0,
    xOffset=0, yOffset=0, zOffset=0, tiltX=0, tiltY=0,
    fpScale=1,
    b=None, R=None
):
    """Convert xyz wok coordinates in mm to xyz focal coordinates in mm.

    The origin of the focal coordinate system is the
    M1 vertex. focal +y points toward North, +x points toward E.
    The origin of the wok coordinate system is the wok vertex.  -x points
    toward the boss slithead.  +z points from telescope to sky.

    Tilt is applied first about x axis then y axis.

    Note: currently zWok is irrelevant using a flat wok model.


    Parameters
    -------------
    xWok: scalar or 1D array
        x position of object in wok space mm
        (+x aligned with +RA on image at PA=0)
    yWok: scalar or 1D array
        y position of object in wok space mm
        (+y aligned with +Dec on image at PA=0)
    zWok: scalar or 1D array
        z position of object in wok space mm
        (+z aligned boresight and increases from the telescope to the sky)
    positionAngle: scalar
        position angle deg.  Angle measured from (image) North through East to wok +y.
        So position angle of 45 deg, wok +y points NE
    xOffset: scalar or None
        x position (mm) of wok origin (vertex) in focal coords
        calibrated
    yOffset: scalar
        y position (mm) of wok origin (vertex) in focal coords
        calibratated
    zOffset: scalar
        z position (mm) of wok origin (vertex) in focal coords
        calibratated
    tiltX: scalar
        tilt (deg) of wok about focal x axis at PA=0
        calibrated
    tiltY: scalar
        tilt (deg) of wok about focal y axis at PA=0
        calibrated
    fpScale: scalar
        radial scale multiplier
    b: float
        z location of the center of curvature, dist from M1 vertex (mm)
    R: float
        radius of curvature (mm)


    Returns
    ---------
    xFocal: scalar or 1D array
        x position of object in focal coord sys mm
        (+x aligned with +RA on image)
    yFocal: scalar or 1D array
        y position of object in focal coord sys mm
        (+y aligned with +Dec on image)
    zFocal: scalar or 1D array
        z position of object in focal coord sys mm
        (+z aligned boresight and increases from the telescope to the sky)
    """

    # scale xyWok from dither analysis results
    xWok = numpy.array(xWok) * fpScale # from gfa: 0.9998899
    yWok = numpy.array(yWok) * fpScale # from gfa: 0.9998899

    # this routine is a reversal of the steps
    # in the function focalToWok, with rotational
    # angles inverted and translations applied in reverse
    coords = numpy.array([xWok, yWok, zWok])

    rotX = numpy.radians(-1*tiltX)
    rotY = numpy.radians(-1*tiltY)
    rotZ = numpy.radians(positionAngle)

    rotMatX = numpy.array([
        [1, 0, 0],
        [0, numpy.cos(rotX), numpy.sin(rotX)],
        [0, -numpy.sin(rotX), numpy.cos(rotX)]
    ])

    rotMatY = numpy.array([
        [numpy.cos(rotY), 0, -numpy.sin(rotY)],
        [0, 1, 0],
        [numpy.sin(rotY), 0, numpy.cos(rotY)]
    ])

    rotMatZ = numpy.array([
        [numpy.cos(rotZ), numpy.sin(rotZ), 0],
        [-numpy.sin(rotZ), numpy.cos(rotZ), 0],
        [0, 0, 1]
    ])

    transXYZ = numpy.array([xOffset, yOffset, zOffset])

    # add offsets for reverse transform
    if hasattr(xWok, "__len__"):
        # list of coords fed in
        transXYZ = numpy.array([transXYZ]*len(xWok)).T
        coords = coords + transXYZ
    else:
        # single set of xyz coords fed in
        coords = coords + transXYZ

    coords = rotMatZ.dot(coords)
    coords = rotMatY.dot(coords)
    xFocal, yFocal, zFocal = rotMatX.dot(coords)

    if b is not None or R is not None:
        # calculate z coords based on xy
        # basically force the point to land on the focal
        # surface for a given wavelength
        # necessary for a flat wok model
        # and this should be removed if we go
        # to curved wok
        # !!!!!!!!!!!!!!!!!
        if R is None or b is None:
            raise CoordIOError("Must specify both b and R if they are not None")
        rFocal = numpy.sqrt(xFocal**2+yFocal**2)
        zFocal = -1*numpy.sqrt(R**2-rFocal**2) + b

    return xFocal, yFocal, zFocal


def _verify3Vector(checkMe, label):
    if not hasattr(checkMe, "__len__"):
        raise RuntimeError("%s must be a 3-vector"%label)
    checkMe = numpy.array(checkMe).squeeze()
    if len(checkMe) != 3:
        raise RuntimeError("%s must be a 3-vector"%label)
    if checkMe.ndim != 1:
        raise RuntimeError("%s must be a 3-vector"%label)
    return checkMe


def repeat_scalar(input, n):
    """Returns an array with an scalar repeated ``n`` times. Does not affect arrays."""

    if numpy.isscalar(input):
        return numpy.repeat(input, n)

    return input


def wokToTangent(xWok, yWok, zWok, b, iHat, jHat, kHat,
                 elementHeight=defaults.POSITIONER_HEIGHT, scaleFac=1,
                 dx=0, dy=0, dz=0):
    """
    Convert an array of wok coordinates to tangent coordinates.

    xyz Wok coords are mm with orgin at wok vertex. +z points from wok toward M2.
    -x points toward the boss slithead

    In the tangent coordinate frame the xy plane is tangent to the wok
    surface at xyz position b.  The origin is set to be elementHeight above
    the wok surface.

    Parameters
    -------------
    xWok: 1D array
        x position of object in wok space mm
        (+x aligned with +RA on image at PA = 0)
    yWok: 1D array
        y position of object in wok space mm
        (+y aligned with +Dec on image at PA = 0)
    zWok: 1D array
        z position of object in wok space mm
        (+z aligned boresight and increases from the telescope to the sky)
    b: 3-element 1D vector
        x,y,z position (mm) of each hole element on wok surface measured in wok coords
    iHat: 3-element 1D vector
        x,y,z unit vector in wok coords that indicate the direction
        of the tangent coordinate x axis for each hole.
    jHat: 3-element 1D vector
        x,y,z unit vector in wok coords that indicate the direction
        of the tangent coordinate y axis for each hole.
    kHat: 3-element 1D vector
        x,y,z unit vector in wok coords that indicate the direction
        of the tangent coordinate z axis for each hole.
    elementHeight: 1D array
        height (mm) of positioner/GFA chip above wok surface
    scaleFac: 1D array
        scale factor to apply to b to account for thermal expansion of wok.
        scale is applied to b radially
    dx: 1D array
        x offset (mm), calibration to capture small displacements of
        tangent x
    dy: 1D array
        y offset (mm), calibration to capture small displacements of
        tangent y
    dz: 1D array
        z offset (mm), calibration to capture small displacements of
        tangent x

    Returns
    ---------
    xTangent: 1D array
        x position (mm) in tangent coordinates
    yTangent: 1D array
        y position (mm) in tangent coordinates
    zTangent: 1D array
        z position (mm) in tangent coordinates
    """

    b = _verify3Vector(b, 'b').tolist()
    iHat = _verify3Vector(iHat, 'iHat').tolist()
    jHat = _verify3Vector(jHat, 'jHat').tolist()
    kHat = _verify3Vector(kHat, 'kHat').tolist()

    xyzWok = numpy.atleast_2d(numpy.array([xWok, yWok, zWok]).T).tolist()

    xyzTangent = libcoordio.wokToTangentArr(
        xyzWok, b, iHat, jHat, kHat,
        elementHeight, scaleFac, dx, dy, dz
    )

    xyzTangent = numpy.array(xyzTangent)
    xTangent = xyzTangent[:, 0]
    yTangent = xyzTangent[:, 1]
    zTangent = xyzTangent[:, 2]

    return xTangent, yTangent, zTangent


def _wokToTangent(xWok, yWok, zWok, b, iHat, jHat, kHat,
                  elementHeight=defaults.POSITIONER_HEIGHT, scaleFac=1,
                  dx=0, dy=0, dz=0):
    """
    Convert from wok coordinates to tangent coordinates.

    xyz Wok coords are mm with orgin at wok vertex. +z points from wok toward M2.
    -x points toward the boss slithead

    In the tangent coordinate frame the xy plane is tangent to the wok
    surface at xyz position b.  The origin is set to be elementHeight above
    the wok surface.

    Parameters
    -------------
    xWok: scalar or 1D array
        x position of object in wok space mm
        (+x aligned with +RA on image at PA = 0)
    yWok: scalar or 1D array
        y position of object in wok space mm
        (+y aligned with +Dec on image at PA = 0)
    zWok: scalar or 1D array
        z position of object in wok space mm
        (+z aligned boresight and increases from the telescope to the sky)
    b: 3-vector
        x,y,z position (mm) of element hole on wok surface measured in wok coords
    iHat: 3-vector
        x,y,z unit vector in wok coords that indicate the direction
        of the tangent coordinate x axis
    jHat: 3-vector
        x,y,z unit vector in wok coords that indicate the direction
        of the tangent coordinate y axis
    kHat: 3-vector
        x,y,z unit vector in wok coords that indicate the direction
        of the tangent coordinate z axis
    elementHeight: scalar
        height (mm) of positioner/GFA chip above wok surface
    scaleFac: scalar
        scale factor to apply to b to account for thermal expansion of wok.
        scale is applied to b radially
    dx: scalar
        x offset (mm), calibration to capture small displacements of
        tangent x
    dy: scalar
        y offset (mm), calibration to capture small displacements of
        tangent y
    dz: scalar
        z offset (mm), calibration to capture small displacements of
        tangent x

    Returns
    ---------
    xTangent: scalar or 1D array
        x position (mm) in tangent coordinates
    yTangent: scalar or 1D array
        y position (mm) in tangent coordinates
    zTangent: scalar or 1D array
        z position (mm) in tangent coordinates
    """
    # check for problems
    b = _verify3Vector(b, "b")
    iHat = _verify3Vector(iHat, "iHat")
    jHat = _verify3Vector(jHat, "jHat")
    kHat = _verify3Vector(kHat, "kHat")

    coords = numpy.array([xWok,yWok,zWok])

    # apply radial scale factor to b
    # assume that z is unaffected by
    # thermal expansion (probably reasonable)
    if scaleFac != 1:
        r = numpy.sqrt(b[0]**2+b[1]**2)
        theta = numpy.arctan2(b[1], b[0])
        r = r*scaleFac
        b[0] = r*numpy.cos(theta)
        b[1] = r*numpy.sin(theta)

    isArr = hasattr(xWok, "__len__")
    calibOff = numpy.array([dx,dy,dz])
    if isArr:
        b = numpy.array([b]*len(xWok)).T
        calibOff = numpy.array([calibOff]*len(xWok)).T

    # offset to wok base position
    coords = coords - b

    # rotate normal to wok surface at point b
    rotNorm = numpy.array([iHat, jHat, kHat], dtype="float64")
    coords = rotNorm.dot(coords)

    # offset xy plane to focal surface
    if isArr:
        coords[2,:] = coords[2,:] - elementHeight
    else:
        coords[2] = coords[2] - elementHeight

    # apply rotational calibration
    # if dRot != 0:
    #     rotZ = numpy.radians(dRot)
    #     rotMatZ = numpy.array([
    #         [numpy.cos(rotZ), numpy.sin(rotZ), 0],
    #         [-numpy.sin(rotZ), numpy.cos(rotZ), 0],
    #         [0, 0, 1]
    #     ])

    #     coords = rotMatZ.dot(coords)

    coords = coords - calibOff

    return coords[0], coords[1], coords[2]


def tangentToWok(xTangent, yTangent, zTangent, b, iHat, jHat, kHat,
                 elementHeight=defaults.POSITIONER_HEIGHT, scaleFac=1,
                 dx=0, dy=0, dz=0):
    """
    Convert from tangent coordinates at b to wok coordinates.

    xyz Wok coords are mm with origin at wok vertex. +z points from wok toward M2.
    -x points toward the boss slithead

    In the tangent coordinate frame the xy plane is tangent to the wok
    surface at xyz position b.  The origin is set to be elementHeight above
    the wok surface.

    Parameters
    -------------
    xTangent: scalar or 1D array
        x position (mm) in tangent coordinates
    yTangent: scalar or 1D array
        y position (mm) in tangent coordinates
    zTangent: scalar or 1D array
        z position (mm) in tangent coordinates
    b: 3-vector
        x,y,z position (mm) of element hole on wok surface measured in wok coords
    iHat: 3-vector
        x,y,z unit vector in wok coords that indicate the direction
        of the tangent coordinate x axis
    jHat: 3-vector
        x,y,z unit vector in wok coords that indicate the direction
        of the tangent coordinate y axis
    kHat: 3-vector
        x,y,z unit vector in wok coords that indicate the direction
        of the tangent coordinate z axis
    elementHeight: scalar
        height (mm) of positioner/GFA chip above wok surface
    scaleFac: scalar
        scale factor to apply to b to account for thermal expansion of wok.
        scale is applied to b radially
    dx: scalar
        x offset (mm), calibration to capture small displacements of
        tangent x
    dy: scalar
        y offset (mm), calibration to capture small displacements of
        tangent y
    dz: scalar
        z offset (mm), calibration to capture small displacements of
        tangent x

    Returns
    ---------
    xWok: scalar or 1D array
        x position of object in wok space mm
        (+x aligned with +RA on image at PA = 0)
    yWok: scalar or 1D array
        y position of object in wok space mm
        (+y aligned with +Dec on image at PA = 0)
    zWok: scalar or 1D array
        z position of object in wok space mm
        (+z aligned boresight and increases from the telescope to the sky)
    """
    b = _verify3Vector(b, "b")
    iHat = _verify3Vector(iHat, "iHat")
    jHat = _verify3Vector(jHat, "jHat")
    kHat = _verify3Vector(kHat, "kHat")
    # format into list of lists
    if hasattr(xTangent, "__len__"):
        xyzTangent = numpy.array([xTangent,yTangent,zTangent]).T.tolist()
        xyzWok = libcoordio.tangentToWokArr(
            xyzTangent, list(b), list(iHat), list(jHat), list(kHat),
            elementHeight, scaleFac, dx, dy, dz)

        xyzWok = numpy.array(xyzWok)
        xWok = xyzWok[:,0]
        yWok = xyzWok[:,1]
        zWok = xyzWok[:,2]
    else:
        xWok, yWok, zWok = libcoordio.tangentToWok(
            [xTangent, yTangent, zTangent], list(b), list(iHat), list(jHat),
            list(kHat), elementHeight, scaleFac, dx, dy, dz
        )
    return xWok, yWok, zWok


def _tangentToWok(xTangent, yTangent, zTangent, b, iHat, jHat, kHat,
                 elementHeight=defaults.POSITIONER_HEIGHT, scaleFac=1,
                 dx=0, dy=0, dz=0):
    """
    Convert from tangent coordinates at b to wok coordinates.

    xyz Wok coords are mm with orgin at wok vertex. +z points from wok toward M2.
    -x points toward the boss slithead

    In the tangent coordinate frame the xy plane is tangent to the wok
    surface at xyz position b.  The origin is set to be elementHeight above
    the wok surface.

    Parameters
    -------------
    xTangent: scalar or 1D array
        x position (mm) in tangent coordinates
    yTangent: scalar or 1D array
        y position (mm) in tangent coordinates
    zTangent: scalar or 1D array
        z position (mm) in tangent coordinates
    b: 3-vector
        x,y,z position (mm) of element hole on wok surface measured in wok coords
    iHat: 3-vector
        x,y,z unit vector in wok coords that indicate the direction
        of the tangent coordinate x axis
    jHat: 3-vector
        x,y,z unit vector in wok coords that indicate the direction
        of the tangent coordinate y axis
    kHat: 3-vector
        x,y,z unit vector in wok coords that indicate the direction
        of the tangent coordinate z axis
    elementHeight: scalar
        height (mm) of positioner/GFA chip above wok surface
    scaleFac: scalar
        scale factor to apply to b to account for thermal expansion of wok.
        scale is applied to b radially
    dx: scalar
        x offset (mm), calibration to capture small displacements of
        tangent x
    dy: scalar
        y offset (mm), calibration to capture small displacements of
        tangent y
    dz: scalar
        z offset (mm), calibration to capture small displacements of
        tangent x

    Returns
    ---------
    xWok: scalar or 1D array
        x position of object in wok space mm
        (+x aligned with +RA on image at PA = 0)
    yWok: scalar or 1D array
        y position of object in wok space mm
        (+y aligned with +Dec on image at PA = 0)
    zWok: scalar or 1D array
        z position of object in wok space mm
        (+z aligned boresight and increases from the telescope to the sky)
    """
    # check for problems
    b = _verify3Vector(b, "b")
    iHat = _verify3Vector(iHat, "iHat")
    jHat = _verify3Vector(jHat, "jHat")
    kHat = _verify3Vector(kHat, "kHat")

    calibOff = numpy.array([dx,dy,dz])
    isArr = hasattr(xTangent, "__len__")
    if isArr:
        calibOff = numpy.array([calibOff]*len(xTangent)).T

    coords = numpy.array([xTangent,yTangent,zTangent])

    # apply calibration offsets
    coords = coords + calibOff

    # apply rotational calibration
    # if dRot != 0:
    #     rotZ = -1*numpy.radians(dRot)
    #     rotMatZ = numpy.array([
    #         [numpy.cos(rotZ), numpy.sin(rotZ), 0],
    #         [-numpy.sin(rotZ), numpy.cos(rotZ), 0],
    #         [0, 0, 1]
    #     ])

    #     coords = rotMatZ.dot(coords)

    # offset xy plane by element height
    if isArr:
        coords[2, :] = coords[2, :] + elementHeight
    else:
        coords[2] = coords[2] + elementHeight

    # rotate normal to wok surface at point b
    # invRotNorm = numpy.linalg.inv(numpy.array([iHat, jHat, kHat], dtype="float64"))
    # transpose is inverse! ortho-normal rows or unit vectors!
    invRotNorm = numpy.array([iHat, jHat, kHat], dtype="float64").T

    coords = invRotNorm.dot(coords)

    # offset to wok base position
    # apply radial scale factor to b
    # assume that z is unaffected by
    # thermal expansion (probably reasonable)
    if scaleFac != 1:
        r = numpy.sqrt(b[0]**2+b[1]**2)
        theta = numpy.arctan2(b[1], b[0])
        r = r * scaleFac
        b[0] = r * numpy.cos(theta)
        b[1] = r * numpy.sin(theta)

    if isArr:
        b = numpy.array([b]*len(xTangent)).T
    coords = coords + b

    return coords[0], coords[1], coords[2]


def proj2XYplane(x, y, z, rayOrigin):
    """Given a point x, y, z orginating from rayOrigin, project
    this point onto the xy plane.

    Parameters
    ------------
    x: scalar or 1D array
        x position of point to be projected
    y: scalar or 1D array
        y position of point to be projected
    z: scalar or 1D array
        z position of point to be projected
    rayOrigin: 3-vector
        [x,y,z] origin of ray(s)

    Returns
    --------
    xProj: scalar or 1D array
        projected x value
    yProj: scalar or 1D array
        projected y value
    zProj: scalar or 1D array
        projected z value (should be zero always!)
    projDist: scalar or 1D array
        projected distance (a proxy for something like focus offset)
    """

    # intersections of lines to planes...
    # http://geomalgorithms.com/a05-_intersect-1.html
    # intersect3D_SegmentPlane
    rayOrigin = _verify3Vector(rayOrigin, "rayOrigin")
    if hasattr(x, "__len__"):
        rayOrigin = numpy.array([rayOrigin] * len(z), dtype="float64").T
        x = numpy.array(x, dtype="float64")
        y = numpy.array(y, dtype="float64")
        z = numpy.array(z, dtype="float64")

    xyz = numpy.array([x, y, z], dtype="float64")

    u = xyz - rayOrigin
    w = rayOrigin
    zHat = numpy.array([0, 0, 1])  # normal to xy plane
    D = zHat.dot(u)
    N = -1 * zHat.dot(w)

    sI = N / D
    xyzProj = rayOrigin + sI * u
    projDist = numpy.linalg.norm(xyz - xyzProj, axis=0)
    # negative projDist for points below xy plane
    projDist *= numpy.sign(z)

    return xyzProj[0], xyzProj[1], xyzProj[2], projDist

# def tangentToPositioner2(
#     xTangent, yTangent, xBeta, yBeta, la=7.4, alphaOffDeg=0, betaOffDeg=0
# ):
#     """
#     Determine alpha/beta positioner angles that place xBeta, yBeta coords in mm
#     at xTangent, yTangent.  New CPP Version.  Should never return NaN values
#     for alpha beta, and instead.

#     note: vectorized option not here yet.

#     todo: include hooks for positioner non-linearity

#     Parameters
#     -------------
#     xTangent: float
#         x position (mm) in tangent coordinates
#     yTangent: float
#         y position (mm) in tangent coordinates
#     xBeta: float
#         x position (mm) in beta arm frame
#     yBeta: float
#         y position (mm) in beta arm frame
#     la: float
#         length (mm) of alpha arm
#     alphaOffDeg: float
#         alpha zeropoint offset (deg)
#     betaOffDeg: float
#         beta zeropoint offset (deg)

#     Returns
#     ---------
#     alphaDegRH: scalar or 1D array
#         alpha angle in degrees for a right-handed robot
#     betaDegRH: scalar or 1D array
#         beta angle in degrees for a right-handed robot
#     alphaDegLH: scalar or 1D array
#         alpha angle in degrees for a right-lefthanded robot
#     betaDegLH: scalar or 1D array
#         beta angle in degrees for a left-handed robot
#     err: scalar or 1D array
#         distance beween fiber and xyWok.  This is non-zero when
#         xyWok is outside the positioners workspace
#     """

#     # C++ wrapped stuff wants lists, not numpy arrays
#         alphaRH, betaRH, alphaLH, betaLH, dist = libcoordio.tangentToPositioner2(
#             [xTangent, yTangent], [xBeta, yBeta],
#             la, alphaOffDeg, betaOffDeg
#         )

#     return alphaRH, betaRH, alphaLH, betaLH, dist


def tangentToPositioner(
    xTangent, yTangent, xBeta, yBeta, la=7.4, alphaOffDeg=0, betaOffDeg=0,
    lefthand=False
):
    """
    Determine alpha/beta positioner angles that place xBeta, yBeta coords in mm
    at xTangent, yTangent.  CPP Version.

    todo: include hooks for positioner non-linearity

    Parameters
    -------------
    xTangent: scalar or 1D array
        x position (mm) in tangent coordinates
    yTangent: scalar or 1D array
        y position (mm) in tangent coordinates
    xBeta: scalar or 1D array
        x position (mm) in beta arm frame
    yBeta: scalar or 1D array
        y position (mm) in beta arm frame
    la: scalar or 1D array
        length (mm) of alpha arm
    alphaOffDeg: scalar
        alpha zeropoint offset (deg)
    betaOffDeg: scalar
        beta zeropoint offset (deg)
    lefthand: boolean
        whether to return a lefthand or righthand parity robot

    Returns
    ---------
    alphaDeg: scalar or 1D array
        alpha angle in degrees
    betaDeg: scalar or 1D array
        beta angle in degrees
    isOK: boolean or 1D boolean array
        True if point physically accessible, False otherwise
    """
    # C++ wrapped stuff wants lists, not numpy arrays
    isArr = hasattr(xTangent, "__len__")
    if isArr:
        xyTangent = numpy.array([xTangent, yTangent]).T.tolist()

        if hasattr(xBeta, "__len__"):
            xyBeta = numpy.array([xBeta, yBeta]).T.tolist()
        else:
            xyBeta = [[xBeta, yBeta]]*len(xyTangent)

        alphaBeta = libcoordio.tangentToPositionerArr(
            xyTangent, xyBeta, la, alphaOffDeg, betaOffDeg, lefthand
        )

        alphaBeta = numpy.array(alphaBeta)
        alpha = alphaBeta[:, 0]
        beta = alphaBeta[:, 1]
    else:
        alpha, beta = libcoordio.tangentToPositioner(
            [xTangent, yTangent], [xBeta, yBeta],
            la, alphaOffDeg, betaOffDeg, lefthand
        )

    isOKAlpha = numpy.isfinite(alpha)
    isOKBeta = numpy.isfinite(beta)
    isOK = isOKAlpha & isOKBeta

    if not isArr:
        isOK = bool(isOK)

    return alpha, beta, isOK


def _tangentToPositioner(
    xTangent, yTangent, xBeta, yBeta, la=7.4, alphaOffDeg=0, betaOffDeg=0
):
    """
    Determine alpha/beta positioner angles that place xBeta, yBeta coords in mm
    at xTangent, yTangent.  Python Version.

    todo: include hooks for positioner non-linearity

    note, didn't add left hand option here...use C++

    Parameters
    -------------
    xTangent: scalar or 1D array
        x position (mm) in tangent coordinates
    yTangent: scalar or 1D array
        y position (mm) in tangent coordinates
    xBeta: scalar or 1D array
        x position (mm) in beta arm frame
    yBeta: scalar or 1D array
        y position (mm) in beta arm frame
    la: scalar or 1D array
        length (mm) of alpha arm
    alphaOffDeg: scalar
        alpha zeropoint offset (deg)
    betaOffDeg: scalar
        beta zeropoint offset (deg)

    Returns
    ---------
    alphaDeg: scalar or 1D array
        alpha angle in degrees
    betaDeg: scalar or 1D array
        beta angle in degrees
    isOK: boolean or 1D boolean array
        True if point physically accessible, False otherwise
    """

    isArr = hasattr(xTangent, "__len__")
    if isArr:
        xTangent = numpy.array(xTangent, dtype="float64")
        yTangent = numpy.array(yTangent, dtype="float64")

    if hasattr(xBeta, "__len__"):
        xBeta = numpy.array(xBeta, dtype="float64")
        yBeta = numpy.array(yBeta, dtype="float64")

    # polar coords jive better for this calculation
    thetaTangent = numpy.arctan2(yTangent, xTangent)
    rTangentSq = xTangent**2 + yTangent**2

    # convert xy Beta to radial coords
    # the origin of the beta coord system is the
    # beta axis of rotation
    thetaBAC = numpy.arctan2(yBeta, xBeta) # radians!
    rBacSq = xBeta**2+yBeta**2

    gamma = numpy.arccos(
        (la**2 + rBacSq - rTangentSq) / (2 * la * numpy.sqrt(rBacSq))
    )
    xi = numpy.arccos(
        (la**2 + rTangentSq - rBacSq) / (2 * la * numpy.sqrt(rTangentSq))
    )

    thetaTangent = numpy.degrees(thetaTangent)
    thetaBAC = numpy.degrees(thetaBAC)
    gamma = numpy.degrees(gamma)
    xi = numpy.degrees(xi)

    alphaDeg = thetaTangent - xi - alphaOffDeg
    betaDeg = 180 - gamma - thetaBAC - betaOffDeg

    # look for nans
    isOKAlpha = numpy.isfinite(alphaDeg)
    isOKBeta = numpy.isfinite(betaDeg)
    isOK = isOKAlpha & isOKBeta

    if not isArr:
        isOK = bool(isOK)

    # handle wrapping? not sure it's necessary
    alphaDeg = alphaDeg % 360
    betaDeg = betaDeg % 360  # should already be there, but...
    # if hasattr(betaDeg, "__len__"):
    #     if True in betaDeg > 180:
    #         raise RuntimeError("problem in alpha/beta conversion! beta > 180")
    # elif betaDeg > 180:
    #     raise RuntimeError("problem in alpha/beta conversion! beta > 180")

    return alphaDeg, betaDeg, isOK


def positionerToTangent(
    alphaDeg, betaDeg, xBeta, yBeta, la=7.4, alphaOffDeg=0, betaOffDeg=0
):
    """
    Determine tangent coordinates (mm) of xBeta, yBeta coords in mm
    from alpha/beta angle.  CPP version.

    todo: include hooks for positioner non-linearity

    Parameters
    -------------
    alphaDeg: scalar or 1D array
        alpha angle in degrees
    betaDeg: scalar or 1D array
        beta angle in degrees
    xBeta: scalar or 1D array
        x position (mm) in beta arm frame
    yBeta: scalar or 1D array
        y position (mm) in beta arm frame
    la: scalar or 1D array
        length (mm) of alpha arm
    alphaOffDeg: scalar
        alpha zeropoint offset (deg)
    betaOffDeg: scalar
        beta zeropoint offset (deg)

    Returns
    ---------
    xTangent: scalar or 1D array
        x position (mm) in tangent coordinates
    yTangent: scalar or 1D array
        y position (mm) in tangent coordinates
    """
    if hasattr(alphaDeg, "__len__"):
        alphaBeta = numpy.array([alphaDeg, betaDeg]).T.tolist()

        if hasattr(xBeta, "__len__"):
            xyBeta = numpy.array([xBeta, yBeta]).T.tolist()
        else:
            xyBeta = [[xBeta, yBeta]] * len(alphaBeta)

        xyTangent = libcoordio.positionerToTangentArr(
            alphaBeta, xyBeta, la, alphaOffDeg, betaOffDeg
        )

        xyTangent = numpy.array(xyTangent)
        xTangent = xyTangent[:, 0]
        yTangent = xyTangent[:, 1]
    else:
        xTangent, yTangent = libcoordio.positionerToTangent(
            [alphaDeg, betaDeg], [xBeta, yBeta],
            la, alphaOffDeg, betaOffDeg
        )
    return xTangent, yTangent


def _positionerToTangent(
    alphaDeg, betaDeg, xBeta, yBeta, la=7.4, alphaOffDeg=0, betaOffDeg=0
):
    """
    Determine tangent coordinates (mm) of xBeta, yBeta coords in mm
    from alpha/beta angle.  Python version.

    todo: include hooks for positioner non-linearity

    Parameters
    -------------
    alphaDeg: scalar or 1D array
        alpha angle in degrees
    betaDeg: scalar or 1D array
        beta angle in degrees
    xBeta: scalar or 1D array
        x position (mm) in beta arm frame
    yBeta: scalar or 1D array
        y position (mm) in beta arm frame
    la: scalar or 1D array
        length (mm) of alpha arm
    alphaOffDeg: scalar
        alpha zeropoint offset (deg)
    betaOffDeg: scalar
        beta zeropoint offset (deg)

    Returns
    ---------
    xTangent: scalar or 1D array
        x position (mm) in tangent coordinates
    yTangent: scalar or 1D array
        y position (mm) in tangent coordinates
    """
    # convert xy Beta to radial coords
    # the origin of the beta coord system is the
    # beta axis of rotation
    if hasattr(xBeta, "__len__"):
        xBeta = numpy.array(xBeta, dtype="float64")
        yBeta = numpy.array(yBeta, dtype="float64")

    thetaBAC = numpy.arctan2(yBeta, xBeta)  # radians!
    rBAC = numpy.sqrt(xBeta**2 + yBeta**2)
    alphaRad = numpy.radians(alphaDeg+alphaOffDeg)
    betaRad = numpy.radians(betaDeg+betaOffDeg)

    cosAlpha = numpy.cos(alphaRad)
    sinAlpha = numpy.sin(alphaRad)
    cosAlphaBeta = numpy.cos(alphaRad + betaRad + thetaBAC)
    sinAlphaBeta = numpy.sin(alphaRad + betaRad + thetaBAC)

    xTangent = la * cosAlpha + rBAC * cosAlphaBeta
    yTangent = la * sinAlpha + rBAC * sinAlphaBeta

    return xTangent, yTangent


def tangentToGuide(xTangent, yTangent, xBin=1, yBin=1):

    xPix = (1 / xBin) * (
        defaults.MICRONS_PER_MM / defaults.GFA_PIXEL_SIZE * xTangent + \
        defaults.GFA_CHIP_CENTER
    )

    yPix = (1 / yBin) * (
        defaults.MICRONS_PER_MM / defaults.GFA_PIXEL_SIZE * yTangent + \
        defaults.GFA_CHIP_CENTER
    )

    return xPix, yPix


def guideToTangent(xPix, yPix, xBin=1, yBin=1):
    xTangent = (
        (xPix * xBin - defaults.GFA_CHIP_CENTER)
        * defaults.GFA_PIXEL_SIZE
        / defaults.MICRONS_PER_MM
    )

    yTangent = (
        (yPix * yBin - defaults.GFA_CHIP_CENTER)
        * defaults.GFA_PIXEL_SIZE
        / defaults.MICRONS_PER_MM
    )

    return xTangent, yTangent


# class FVCUW(object):
#     # right now this is for UWs test bench, fix image scale and use
#     # euclidean transforms
#     defaultScale = 0.026926930620282834
#     defaultRotation = -0.005353542811111436
#     defaultTranslation = numpy.array([-75.69606047, -48.68561274])
#     # transform from CCD pixels to mm
#     def __init__(self):

#         self.tform = SimilarityTransform(
#             scale=self.defaultScale,
#             rotation=self.defaultRotation,
#             translation=self.defaultTranslation
#          )

#     def fit(self, xpix, ypix, xWok, yWok):
#         xypixels = numpy.array([xpix, ypix]).T
#         xyWok = numpy.array([xWok,yWok]).T
#         self.tform.estimate(xypixels, xyWok)


#     def fvcToWok(self, xpix, ypix):
#         xypixels = numpy.array([xpix, ypix]).T
#         xyWok = self.tform(xypixels)
#         return xyWok[:,0], xyWok[:,1]

#     def wokToFVC(self, xWok, yWok):
#         xyWok = numpy.array([xWok, yWok]).T
#         xypix = self.tform.inverse(xyWok)
#         return xypix[:,0], xypix[:,1]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-09-13
# @Filename: astrometry.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import os
import pathlib
import subprocess
import tempfile
import time
from typing import NamedTuple, Optional, TypeVar

import pandas
from astropy.table import Table
from astropy.wcs import WCS


PathLike = TypeVar("PathLike", pathlib.Path, str)


class TimedProcess(NamedTuple):
    """A completed process which includes its elapsed time."""

    process: asyncio.subprocess.Process | subprocess.CompletedProcess
    elapsed: float


class AstrometryNet:
    """A wrapper for the astrometry.net ``solve-field`` command.

    Parameters
    ----------
    configure_params
        Parameters to be passed to `.configure`.
    """

    def __init__(self, **configure_params):
        solve_field_cmd = subprocess.run(
            "which solve-field",
            shell=True,
            capture_output=True,
        )
        solve_field_cmd.check_returncode()

        self.solve_field_cmd = solve_field_cmd.stdout.decode().strip()

        self._options = {}
        self.configure(**configure_params)

    def configure(
        self,
        backend_config: Optional[str] = None,
        width: Optional[int] = None,
        height: Optional[int] = None,
        sort_column: Optional[str] = None,
        sort_ascending: Optional[bool] = None,
        no_plots: Optional[bool] = None,
        ra: Optional[float] = None,
        dec: Optional[float] = None,
        radius: Optional[float] = None,
        scale_low: Optional[float] = None,
        scale_high: Optional[float] = None,
        scale_units: Optional[str] = None,
        dir: Optional[str] = None,
        **kwargs,
    ):
        """Configures how to run of ``solve-field```.

        The parameters this method accepts are identical to those of
        ``solve-field`` and are passed unchanged.

        Parameters
        ----------
        backend_config
            Use this config file for the ``astrometry-engine`` program.
        width
            Specify the field width, in pixels.
        height
            Specify the field height, in pixels.
        sort_column
            The FITS column that should be used to sort the sources.
        sort_ascending
            Sort in ascending order (smallest first);
            default is descending order.
        no_plot
            Do not produce plots.
        ra
            RA of field center for search, in degrees.
        dec
            Dec of field center for search, in degrees.
        radius
            Only search in indexes within ``radius`` degrees of the field
            center given by ``ra`` and ``dec``.
        scale_low
            Lower bound of image scale estimate.
        scale_high
            Upper bound of image scale estimate.
        scale_units
            In what units are the lower and upper bounds? Choices:
            ``'degwidth'``, ``'arcminwidth'``, ``'arcsecperpix'``,
            ``'focalmm'``.
        dir
            Path to the directory where all output files will be saved.
        """

        self._options = {
            "backend-config": backend_config,
            "width": width,
            "height": height,
            "sort-column": sort_column,
            "sort-ascending": sort_ascending,
            "no-plots": no_plots,
            "ra": ra,
            "dec": dec,
            "radius": radius,
            "scale-low": scale_low,
            "scale-high": scale_high,
            "scale-units": scale_units,
            "dir": dir,
            "overwrite": True,
        }
        self._options.update(self._format_options(kwargs))

        return

    def _format_options(self, options: dict):
        """Formats the option names."""

        formatted_options = {}

        for key, value in options.items():
            formatted_key = key.replace("_", "-")
            formatted_options[formatted_key] = value

        return formatted_options

    def _build_command(self, files, options=None):
        """Builds the ``solve-field`` command to run."""

        if options is None:
            options = self._options

        flags = ["no-plots", "sort-ascending", "overwrite"]

        cmd = [self.solve_field_cmd]

        for option in options:
            if options[option] is None:
                continue
            if option in flags:
                if options[option] is True:
                    cmd.append("--" + option)
            else:
                cmd.append("--" + option)
                cmd.append(str(options[option]))

        cmd += list(files)

        return cmd

    def run_sync(
        self,
        files: PathLike | list[PathLike],
        shell: bool = True,
        stdout: Optional[PathLike] = None,
        stderr: Optional[PathLike] = None,
        **kwargs,
    ):
        """Runs astrometry.net synchronously.

        Parameters
        ----------
        files
            List of files to be processed.
        shell
            Whether to call `subprocess.run` with ``shell=True``.
        stdout
            Path where to save the stdout output.
        stderr
            Path where to save the stderr output.
        kwargs
            Configuration parameters (see `.configure`) to override. The
            configuration applies only to this run of ``solve-field`` and it
            is not saved.

        Returns
        -------
        `subprocess.CompletedProcess`
            The completed process.

        """

        options = self._options.copy()
        options.update(self._format_options(kwargs))

        if not isinstance(files, (tuple, list)):
            files = [files]

        t0 = time.time()

        args = self._build_command(list(map(str, files)), options=options)
        if shell:
            args = " ".join(args)

        cmd = subprocess.run(
            args,
            shell=shell,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )

        elapsed = time.time() - t0

        if stdout:
            with open(stdout, "wb") as out:
                if isinstance(args, str):
                    out.write(args.encode() + b"\n")
                else:
                    out.write(" ".join(args).encode() + b"\n")
                out.write(cmd.stdout)

        if stderr:
            with open(stderr, "wb") as err:
                err.write(cmd.stderr)

        return TimedProcess(cmd, elapsed)

    async def run_async(
        self,
        files: PathLike | list[PathLike],
        shell: bool = True,
        stdout: Optional[PathLike] = None,
        stderr: Optional[PathLike] = None,
        **kwargs,
    ) -> TimedProcess:
        """Runs astrometry.net asynchronously.

        Parameters
        ----------
        files
            List of files to be processed.
        shell
            Whether to call `subprocess.run` with ``shell=True``.
        stdout
            Path where to save the stdout output.
        stderr
            Path where to save the stderr output.
        kwargs
            Configuration parameters (see `.configure`) to override. The
            configuration applies only to this run of ``solve-field`` and it
            is not saved.

        Returns
        -------
        `subprocess.CompletedProcess`
            The completed process.

        """

        options = self._options.copy()
        options.update(self._format_options(kwargs))

        if not isinstance(files, (tuple, list)):
            files = [files]

        t0 = time.time()

        args = self._build_command(list(map(str, files)), options=options)

        if shell:
            cmd = await asyncio.create_subprocess_shell(
                " ".join(args),
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )

        else:
            cmd = await asyncio.create_subprocess_exec(
                *args,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )

        stdout_bytes, stderr_bytes = await cmd.communicate()

        elapsed = time.time() - t0

        if stdout:
            with open(stdout, "wb") as out:
                if isinstance(args, str):
                    out.write(args.encode() + b"\n")
                else:
                    out.write(" ".join(args).encode() + b"\n")
                out.write(stdout_bytes)

        if stderr:
            with open(stderr, "wb") as err:
                err.write(stderr_bytes)

        return TimedProcess(cmd, elapsed)


def astrometrynet_quick(
    regions: pandas.DataFrame,
    ra: float,
    dec: float,
    pixel_scale: float,
    pixel_scale_factor_hi: float = 1.1,
    pixel_scale_factor_lo: float = 0.9,
    radius: float = 0.5,
    width: float = 2048,
    height: float = 2048,
    index_path: str | None = None,
    verbose: bool = False,
    **kwargs,
):
    """Quickly process a set of detections using astrometry.net.

    Parameters
    ----------
    regions
        A pandas data frame of source detections with at least three columns:
        ``x``, ``y``, and ``flux``. ``flux`` can be set to all ones if the
        fluxes have not been measured.
    ra
        The estimated RA of the field, in degrees.
    dec
        The estimated declination of the field, in degrees.
    pixel_scale
        The estimated pixel scale, in arcsec.
    pixel_scale_factor_lo
        A multiplicative factor for the pixel scale to calculate the
        lowest possible pixel scale to attempt.
    pixel_scale_factor_hi
        A multiplicative factor for the pixel scale to calculate the
        highest possible pixel scale to attempt.
    radius
        The radius, in degrees, around ``ra`` and ``dec`` to search.
    width
        The width of the image in pixels.
    height
        The height of the image in pixels.
    index_path
        The path to the index files to use. By default, uses whatever is
        defined in ``etc/astrometrynet.cfg`` in the coordio package.
    verbose
        Raise an error if the field is not solved.
    kwargs
        Other options to pass to `.AstrometryNet`.

    Returns
    -------
    wcs
        An astropy WCS with the solved field, or `None` if the field could
        not be solved.

    """

    if not isinstance(regions, pandas.DataFrame):
        regions = pandas.DataFrame(regions)

    outtempfile = tempfile.NamedTemporaryFile().name

    backend_config: str
    if index_path:
        index_path = os.path.abspath(index_path)
        backend_config = os.path.join(os.path.dirname(outtempfile), "astrometrynet.cfg")
        with open(backend_config, "w") as ff:
            ff.write(
                f"""inparallel
cpulimit 30
add_path {index_path}
autoindex
"""
            )
    else:
        backend_config = str(pathlib.Path(__file__).parent / "etc/astrometrynet.cfg")

    opts = dict(
        backend_config=str(backend_config),
        width=width,
        height=height,
        no_plots=True,
        scale_low=pixel_scale * pixel_scale_factor_lo,
        scale_high=pixel_scale * pixel_scale_factor_hi,
        scale_units="arcsecperpix",
        sort_column="flux",
        radius=radius,
    )
    opts.update(kwargs)

    astrometry = AstrometryNet(**opts)

    xyls = Table.from_pandas(regions)
    xyls_path = outtempfile + ".xyls"
    xyls.write(xyls_path, format="fits", overwrite=True)

    wcs_output = pathlib.Path(outtempfile + ".wcs")
    wcs_output.unlink(missing_ok=True)

    astrometry.run_sync(
        [xyls_path],
        stdout=outtempfile + ".stdout",
        stderr=outtempfile + ".stderr",
        ra=ra,
        dec=dec,
    )

    if wcs_output.exists():
        wcs = WCS(open(wcs_output).read())
        wcs_output.unlink()
        return wcs
    else:
        if verbose:
            if os.path.exists(outtempfile + ".stdout"):
                stdout = open(outtempfile + ".stdout").read()
                stderr = open(outtempfile + ".stderr").read()
                raise RuntimeError(stdout + "\n" + stderr)
            else:
                raise RuntimeError("Unexpected error. No stderr was generated.")

    return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-08-21
# @Filename: test_iers.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import pytest

from coordio import IERS, CoordIOUserWarning


@pytest.fixture()
def clear_iers_instance():
    # Because of Python mangling for double underscores we need to modify
    # the instance like this: https://bit.ly/3hktKEm.

    # Clear before and after the test, so that this also clears in case that previous
    # tests set the instance.

    IERS._IERS__instance = None
    yield
    IERS._IERS__instance = None


def test_iers(tmpdir, clear_iers_instance):

    iers_file = tmpdir / 'finals2000A.data.csv'

    # raise RuntimeError(iers_file)

    with pytest.warns(CoordIOUserWarning) as ww:
        iers = IERS(path=tmpdir)

    assert 'Downloading IERS table from' in str(ww[0].message)
    assert iers is not None
    assert iers.data is not None

    assert iers_file.exists


def test_is_valid(iers_data_path):

    iers = IERS(iers_data_path)
    assert iers.is_valid(2458863.0)


def test_is_valid_out_of_range():

    iers = IERS()

    with pytest.warns(CoordIOUserWarning):
        assert iers.is_valid(10000000) is False


def test_get_delta_ut1_utc():

    iers = IERS()

    # Jan 15th, 2020
    assert iers.get_delta_ut1_utc(2458863.5) == pytest.approx(-0.1799645)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

import os

import matplotlib.patheffects as PathEffects
import matplotlib.pyplot as plt
import numpy
import pandas
import scipy
import sep
from skimage.transform import SimilarityTransform

from .conv import positionerToTangent, tangentToWok, wokToTangent
from .defaults import POSITIONER_HEIGHT, calibration
from .exceptions import CoordIOError
from .libcoordio import tangentToPositioner, tangentToPositioner2
from .utils import refinexy
from .zhaoburge import fitZhaoBurge, getZhaoBurgeXY


# __all__ = ["RoughTransform", "ZhaoBurgeTransform", "FVCTransformAPO"]


def arg_nearest_neighbor(
    xyA: numpy.ndarray,
    xyB: numpy.ndarray,
    atol: float | None = None,
):
    """Finds the nearest neighbour in list B for each target in list A.

    If the distance between the item in A and the closest element in B is greater
    than ``atol``, a match is not returned.

    Parameters
    ----------
    xyA
        The list we want to match.
    xyB
        The reference table.
    atol
        The maximum allowed distance. `None` to not do any distance checking.

    Returns
    -------
    result
        A tuple with the indices in ``A`` that have been matched, the matching index
        in ``B`` for each matched element in ``A``, and the distance from each
        element in ``A`` to the nearest neighbour in ``B`` (regardless of whether
        that distance is greater than ``atol``).

    """

    xyA = numpy.array(xyA)
    xyB = numpy.array(xyB)

    distances = scipy.spatial.distance.cdist(xyA, xyB)

    min_distances = numpy.array([numpy.min(d) for d in distances])
    indexB = numpy.array([numpy.argmin(d) for d in distances])

    if atol is not None:
        good_matches = numpy.where(min_distances < atol)[0]
    else:
        good_matches = numpy.arange(len(indexB))

    return good_matches, indexB[good_matches], min_distances


def wokToPositioner(
    xWok, yWok,
    xBeta, yBeta, la,
    alphaOffDeg, betaOffDeg,
    dx, dy, b, iHat, jHat, kHat,
    newInvKin=True
):
    """
    Find alpha beta coords of a robot given xyWok coords of a robot's fiber

    note: vectorized option not here yet.  Only right hand solutions
    for now.

    Parameters
    ------------
    xWok : float
        x location of fiber in wok coordinates (mm)
    yWok : float
        y location of fiber in wok coordinates (mm)
    xBeta : float
        x location of fiber in beta coords (mm)
    yBeta : numpy.array or float
        y location of fiber in beta coords (mm)
    la : numpy.array or float
        alpha arm length in mm
    alphaOffDeg : numpy.narray or float
        calibrated alpha arm offset in deg
    betaOffDeg : float
        calibrated beta arm offset in deg
    dx : float
        calibrated robot body x offset in wok coords (mm)
    dy : float
        calibrated robot body y offset in wok coords (mm)
    b: numpy.ndarray
        3-element 1D vector
        x,y,z position (mm) of each hole element on wok
        surface measured in wok coords
    iHat: numpy.ndarray
        3-element 1D vector
        x,y,z unit vector in wok coords that indicate the direction
        of the tangent coordinate x axis for each hole.
    jHat: numpy.ndarray
        3-element 1D vector
        x,y,z unit vector in wok coords that indicate the direction
        of the tangent coordinate y axis for each hole.
    kHat: numpy.ndarray
        3-element 1D vector
        x,y,z unit vector in wok coords that indicate the direction
        of the tangent coordinate z axis for each hole.
    newInvKin : bool
        Use new style inverse kinematics solver


    Returns
    ---------
    alpha
        alpha angle of robot in degrees
    beta
        beta angle of robot in degrees

    """
    xt,yt,zt = wokToTangent(
        xWok, yWok, POSITIONER_HEIGHT, b, iHat, jHat, kHat,
        elementHeight=POSITIONER_HEIGHT, scaleFac=1,
        dx=dx, dy=dy, dz=0
    )

    if newInvKin:
        alpha, beta, alphaLH, betaLH, dist = tangentToPositioner2(
            [xt, yt], [xBeta, yBeta], la, alphaOffDeg, betaOffDeg
        )
    else:
        lefthand = False
        alpha, beta, isOK = tangentToPositioner(
            [xt, yt], [xBeta, yBeta], la, alphaOffDeg, betaOffDeg, lefthand
        )
    return alpha, beta


def positionerToWok(
        alphaDeg, betaDeg,
        xBeta, yBeta, la,
        alphaOffDeg, betaOffDeg,
        dx, dy, b, iHat, jHat, kHat
):
    """
    Find xyWok coords of a fiber given alpha beta coords of a robot

    Parameters
    ------------
    alphaDeg : numpy.ndarray or float
        robot alpha angle coordinate in degrees
    betaDeg : numpy.ndarray or float
        robot beta angle coordinate in degrees
    xBeta : numpy.ndarray or float
        x location of fiber in beta coords (mm)
    yBeta : numpy.array or float
        y location of fiber in beta coords (mm)
    la : numpy.array or float
        alpha arm length in mm
    alphaOffDeg : numpy.narray or float
        calibrated alpha arm offset in deg
    betaOffDeg : numpy.ndarray or float
        calibrated beta arm offset in deg
    dx : numpy.ndarray or float
        calibrated robot body x offset in wok coords (mm)
    dy : numpy.ndarray or float
        calibrated robot body y offset in wok coords (mm)
    b: numpy.ndarray
        3-element 1D vector
        x,y,z position (mm) of each hole element on wok
        surface measured in wok coords
    iHat: numpy.ndarray
        3-element 1D vector
        x,y,z unit vector in wok coords that indicate the direction
        of the tangent coordinate x axis for each hole.
    jHat: numpy.ndarray
        3-element 1D vector
        x,y,z unit vector in wok coords that indicate the direction
        of the tangent coordinate y axis for each hole.
    kHat: numpy.ndarray
        3-element 1D vector
        x,y,z unit vector in wok coords that indicate the direction
        of the tangent coordinate z axis for each hole.


    Returns
    ---------
    xWok
        x loaction of fiber in wok coords (mm)
    yWok
        y location of fiber in wok coords (mm)
    zWok
        z location of fiber in wok coords (mm)


    """

    xt, yt = positionerToTangent(
        alphaDeg, betaDeg, xBeta, yBeta,
        la, alphaOffDeg, betaOffDeg
    )

    if hasattr(xt, "__len__"):
        zt = numpy.zeros(len(xt))
    else:
        zt = 0

    xw, yw, zw = tangentToWok(
        xt, yt, zt, b, iHat, jHat, kHat,
        elementHeight=POSITIONER_HEIGHT, scaleFac=1,
        dx=dx, dy=dy, dz=0

    )

    return xw, yw, zw


def plotFVCResults(
    filename,
    xyFitCentroids,
    xyMetFiber=None,
    xyBossFiber=None,
    xyApogeeFiber=None,
    xyFIF=None,
    positionerIDs=None,
    assoc_used=None,
    xyFitCentroidsUnmatched=None,
    xyFiberWarn=None,
    title=None,
    xyFIFLost=None,
):
    """
    Visualize the results of transformation/association.
    All xy units should be MM.

    filename : str
    xyFitCentroids : numpy.ndarray
        nx2 array, locations of centroids transformed to wok space
    xyMetFiber : numpy.ndarray or None
        nx2 array, expected location of Metrology fiber in wok space
    xyBossFiber : numpy.ndarray or None
         nx2 array, expected location of BOSS fiber in wok space
    xyApogeeFiber : numpy.ndarray or None
        nx2 array, expected location of APOGEE fiber in wok space given robot
        arm angles
    xyFIF : numpy.ndarray or None
        mx2 array, expected location of fiducial fibers in wok space
    positionerIDs : list or None
        lenth n list of integers, if supplied, positioner ids are ploted
        near the points for met, ap, and/or boss fibers
    assoc_used : list or None
        [mx2 array, mx2 array] two sets of xy coords to plot lines between
        for seeing who matched to who (usually fiducials)
    xyFitCentroidsUnmatched : numpy.ndarray or None
        centroids detected but not associated with any robot or fiducial
    xyFiberWarn : numpy.ndarray or None
        fibers with abnormally large errors
    title : str or None
        title for the plot
    xyFIFLost : numpy.ndarray or None
        nx2 array, expected locations of fiducial fibers in wok space that weren't matched

    """

    fig = plt.figure(figsize=(8, 8))
    textProps = {
        "ha": "center", "va": "center",
        "fontsize": 0.15, "fontweight": "bold"
    }

    if title:
        plt.title(title)

    plt.plot(
        xyFitCentroids[:, 0],
        xyFitCentroids[:, 1],
        "o",
        ms=4,
        markerfacecolor="None",
        markeredgecolor="tab:red",
        markeredgewidth=1,
        label="Centroid",
    )

    if xyMetFiber is not None:
        plt.plot(
            xyMetFiber[:, 0],
            xyMetFiber[:, 1],
            "x",
            color="tab:blue",
            # markeredgecolor="white",
            # markeredgewidth=0.1,
            alpha=1,
            ms=3,
            label="Expected MET",
        )
        if positionerIDs is not None:
            for (x, y), pid in zip(xyMetFiber, positionerIDs):
                strID = "P" + ("%i"%pid).zfill(4)
                txt = plt.text(x, y, strID, textProps)

    if xyBossFiber is not None:
        plt.plot(
            xyBossFiber[:, 0],
            xyBossFiber[:, 1],
            "x",
            color="black",
            alpha=1,
            ms=3,
            label="Expected BOSS",
        )
        if positionerIDs is not None:
            for (x, y), pid in zip(xyBossFiber, positionerIDs):
                strID = "P" + ("%i"%pid).zfill(4)
                txt = plt.text(x, y, strID, textProps)
                txt.set_path_effects(
                    [PathEffects.withStroke(linewidth=0.08, foreground='w')]
                )

    if xyApogeeFiber is not None:
        plt.plot(
            xyApogeeFiber[:,0],
            xyApogeeFiber[:,1],
            "x",
            color="tab:purple",
            alpha=1,
            ms=3,
            label="Expected APOGEE",
        )
        if positionerIDs is not None:
            for (x, y), pid in zip(xyApogeeFiber, positionerIDs):
                strID = "P" + ("%i"%pid).zfill(4)
                plt.text(x, y, strID, textProps)

    if xyFitCentroidsUnmatched is not None:
        plt.plot(
            xyFitCentroidsUnmatched[:, 0],
            xyFitCentroidsUnmatched[:, 1],
            "^",
            color="tab:orange",
            alpha=1,
            markerfacecolor="None",
            markeredgewidth=1,
            ms=7,
            label="Unmatched Centroid",
        )

    if xyFiberWarn is not None:
        plt.plot(
            xyFiberWarn[:, 0],
            xyFiberWarn[:, 1],
            "v",
            color="tab:cyan",
            alpha=1,
            markerfacecolor="None",
            markeredgewidth=1,
            ms=7,
            label="Fiber Warning",
        )

    # Overplot fiducials
    if xyFIF is not None:
        plt.plot(
            xyFIF[:,0],
            xyFIF[:,1],
            "D",
            ms=6,
            markerfacecolor="None",
            markeredgecolor="cornflowerblue",
            markeredgewidth=1,
            label="Matched FIF",
        )

    if xyFIFLost is not None:
        plt.plot(
            xyFIFLost[:,0],
            xyFIFLost[:,1],
            "D",
            ms=6,
            markerfacecolor="None",
            markeredgecolor="orange",
            markeredgewidth=1,
            label="Unmatched FIF",
        )

    if assoc_used is not None:
        for xy1, xy2 in zip(assoc_used[0], assoc_used[1]):
            plt.plot([xy1[0], xy2[0]], [xy1[1], xy2[1]], "-k")

    plt.axis("equal")
    # plt.legend(bbox_to_anchor=(1.1, 1.05))
    plt.legend(bbox_to_anchor=(1,1), bbox_transform=fig.transFigure)
    # plt.legend()
    plt.xlim([-350, 350])
    plt.ylim([-350, 350])
    plt.xlabel("Wok x (mm)")
    plt.ylabel("Wok y (mm)")
    plt.savefig(filename, dpi=350)
    plt.close()


class RoughTransform(object):
    """Apply a rough transformation between CCD and Wok Coords.

    Simply subtract mean and scale by variance to put coordinate
    sets into roughly the same space.
    """

    def __init__(self, xyCCD, xyWok):
        """
        xyCCD and xyWok do not need to be a mapping, nor even the same length

        Parameters
        -----------
        xyCCD : numpy.ndarray
            Nx2 array of xy ccd coordinates (pixels)
        xyWok :numpy.ndarray
            Mx2 array of xy wok coordinates (mm)
        """

        # scale pixels to mm roughly
        xCCD = xyCCD[:, 0]
        yCCD = xyCCD[:, 1]
        xWok = xyWok[:, 0]
        yWok = xyWok[:, 1]

        self.meanCCDX = numpy.mean(xCCD)
        self.meanCCDY = numpy.mean(yCCD)
        self.stdCCDX = numpy.std(xCCD)
        self.stdCCDY = numpy.std(yCCD)

        self.stdWokX = numpy.std(xWok)
        self.stdWokY = numpy.std(yWok)

    def apply(self, xyCCD):
        """
        Convert xyCCD to xyWok (very roughly)

        Parameters
        ------------
        xyCCD : numpy.ndarray
            Nx2 array of xy ccd coordinates (pixels)

        Result
        -------
        xyWok
            estimated location of input coords in wok space

        """

        xCCD = xyCCD[:, 0]
        yCCD = xyCCD[:, 1]
        roughWokX = (xCCD - self.meanCCDX) / self.stdCCDX * self.stdWokX
        roughWokY = (yCCD - self.meanCCDY) / self.stdCCDY * self.stdWokY

        return numpy.array([roughWokX, roughWokY]).T


class ZhaoBurgeTransform(object):
    """Apply a full Zhao-Burge transformation.


    Residuals from a similarity tranform are fit with Zhao-Burge
    2D vector polynomials.
    """

    def __init__(self, xyCCD, xyWok, polids=None, normFactor=1):
        """ Compute a transformation between xyCCD and xyWok using
        Zhao-Burge polynomials

        xyCCD and xyWok need to be a mapping by index

        Parameters
        ------------
        xyCCD : numpy.ndarray
            Nx2 array of xy ccd coordinates (pixels)
        xyWok :numpy.ndarray
            Nx2 array of xy wok coordinates (mm)
        polids : numpy.ndarray or None
            Mx1 array of integers specifying the ZB basis vectors to use.
            If not passed a default basis vector set is used.
        normFactor : float
            Scale xyWok coordinates by this factor.  Intended
            use is for moving coordinates to unit circle. 330 is a good
            choice, but default is 1 to remain backwards compatible
            (no normalization)

        Attributes
        -------------
        polids : numpy.ndarray
            1D integer array containing basis IDs used for ZB transform
        coeffs : numpy.ndarray
            1D float array containing the fit coefficient values for each
            polid in the ZB transform
        normFactor : float
            normalization factor used
        simTrans : skimage.transform.SimilarityTransform
            similarity transform fit between xyCCD and xyWok
        unbiasedErrs : numpy.ndarray
            Nx1 array containing zb fit residuals (mm) at each
            jacknife iteration.  At each iter a single point is left out of the
            fit, and that point is used to measure predictive error of the
            model.
        unbiasedRMS : float
            RMS value of unbiasedErrors (mm)
        errs : numpy.ndarray
            Nx1 array containing zb fit resuduals (mm) without
            jacknife (fit using all available data)
        rms : float
            RMS value of errs (mm)
        """

        if polids is None:
            self.polids = numpy.array(
                [0, 1, 2, 3, 4, 5, 6, 9, 20, 28, 29], dtype=numpy.int16
            )
        else:
            self.polids = numpy.array(polids, dtype=numpy.int16)

        if normFactor < 1:
            raise CoordIOError("normFactor must >= 1")

        self.normFactor = normFactor

        # First fit a transrotscale model
        self.simTrans = SimilarityTransform()
        self.simTrans.estimate(xyCCD, xyWok)

        # Apply the model to the data
        xySimTransFit = self.simTrans(xyCCD)

        # Use ZB polys to get the rest of the way use leave-one out xverification to
        # estimate "unbiased" errors in fit.
        self.unbiasedErrs = []
        for ii in range(len(xyCCD)):
            _xyWok = xyWok.copy()
            _xySimTransFit = xySimTransFit.copy()
            _xyWok = numpy.delete(_xyWok, ii, axis=0)
            _xySimTransFit = numpy.delete(_xySimTransFit, ii, axis=0)
            fitCheck = numpy.array(xySimTransFit[ii, :]).reshape((1, 2))
            destCheck = numpy.array(xyWok[ii, :]).reshape((1, 2))

            polids, coeffs = fitZhaoBurge(
                _xySimTransFit[:, 0],
                _xySimTransFit[:, 1],
                _xyWok[:, 0],
                _xyWok[:, 1],
                polids=self.polids,
                normFactor=self.normFactor
            )

            dx, dy = getZhaoBurgeXY(
                polids,
                coeffs,
                fitCheck[:, 0],
                fitCheck[:, 1],
                normFactor=self.normFactor
            )
            zxfit = fitCheck[:, 0] + dx
            zyfit = fitCheck[:, 1] + dy
            zxyfit = numpy.array([zxfit, zyfit]).T
            self.unbiasedErrs.append(destCheck.squeeze() - zxyfit.squeeze())

        self.unbiasedErrs = numpy.array(self.unbiasedErrs)
        self.unbiasedRMS = numpy.sqrt(numpy.mean(self.unbiasedErrs ** 2))

        # Now do the "official fit", using all points

        polids, self.coeffs = fitZhaoBurge(
            xySimTransFit[:, 0],
            xySimTransFit[:, 1],
            xyWok[:, 0],
            xyWok[:, 1],
            polids=self.polids,
            normFactor=self.normFactor
        )

        dx, dy = getZhaoBurgeXY(
            polids,
            self.coeffs,
            xySimTransFit[:, 0],
            xySimTransFit[:, 1],
            normFactor=self.normFactor
        )

        xWokFit = xySimTransFit[:, 0] + dx
        yWokFit = xySimTransFit[:, 1] + dy
        xyWokFit = numpy.array([xWokFit, yWokFit]).T

        self.xyWokIn = xyWok
        self.errs = xyWok - xyWokFit
        self.rms = numpy.sqrt(numpy.mean(self.errs ** 2))

    def apply(self, xyCCD, zb=True):
        """Apply the transformation to a set of (x, y) coordinates.


        Parameters
        -----------
        xyCCD : numpy.ndarray
            Nx2 array of xy ccd coordinates (pixels)
        zb : bool
            If False, only apply the similarity transform
            (no Zhao-Burge higher order components)


        Returns
        ---------
        xyWok
            Nx2 array of wok coordinates
        """

        xySimTransFit = self.simTrans(xyCCD)

        if zb:
            dx, dy = getZhaoBurgeXY(
                self.polids,
                self.coeffs,
                xySimTransFit[:, 0],
                xySimTransFit[:, 1],
                normFactor=self.normFactor
            )
            xWokFit = xySimTransFit[:, 0] + dx
            yWokFit = xySimTransFit[:, 1] + dy
            xyWokFit = numpy.array([xWokFit, yWokFit]).T
        else:
            xyWokFit = xySimTransFit

        return xyWokFit


def xyWokFiberFromPositioner(
        fullTable,
        angleType="Report",
        doMetrology=True,
        doApogee=True,
        doBoss=True
    ):
    """
    Determine xy wok position for a each fiber for each robot.

    Parameters
    ------------
    fullTable : pandas.DataFrame
        A merge on holeID of positionerTable and wokCoords DataFrames,
        with additional columns names specified by alphaColumn and betaColumn
        indicating the robot's arm coords.
    angleType : str
        look for columns alphaXXX betaXXX where XXX=column type
    doMetrology : bool
        if True, compute metrology fiber location
    doApogee : bool
        if True, compute apogee fiber location
    doBoss : bool
        if True, compute boss fiber location

    Returns
    --------
    fullTable : pandas.DataFrame
        the input fullTable with new columns appended
        xWokxxxXXX, yWokxxxXXX where xxx=angleType and
        XXX is one of ["Metrology", "APOGEE","BOSS"]
    """

    # probably can vectorize this whole thing, or at least
    # partially

    fiberNameMap = zip(
        [doMetrology, doApogee, doBoss],
        ["Metrology", "APOGEE", "BOSS"],
        ["met", "ap", "boss"]
    )
    for doFiber, fiberType, colName in fiberNameMap:
        if not doFiber:
            continue
        fiberX = "%sX"%colName
        fiberY = "%sY"%colName
        xWok = []
        yWok = []
        for ii, posRow in fullTable.iterrows():
            alpha = float(posRow["alpha" + angleType])
            beta = float(posRow["beta" + angleType])

            b = numpy.array([posRow.xWok, posRow.yWok, posRow.zWok])
            iHat = numpy.array([posRow.ix, posRow.iy, posRow.iz])
            jHat = numpy.array([posRow.jx, posRow.jy, posRow.jz])
            kHat = numpy.array([posRow.kx, posRow.ky, posRow.kz])
            la = float(posRow.alphaArmLen)
            alphaOffDeg = float(posRow.alphaOffset)
            betaOffDeg = float(posRow.betaOffset)
            dx = float(posRow.dx)
            dy = float(posRow.dy)

            xBeta = float(posRow[fiberX])
            yBeta = float(posRow[fiberY])

            xw, yw, zw = positionerToWok(
                alpha, beta, xBeta, yBeta, la,
                alphaOffDeg, betaOffDeg,
                dx, dy, b, iHat, jHat, kHat
            )

            xWok.append(xw)
            yWok.append(yw)

        fullTable["xWok%s%s" % (angleType, fiberType)] = xWok
        fullTable["yWok%s%s" % (angleType, fiberType)] = yWok

    return fullTable


def alphaBetaFromMetMeas(fullTable, newInvKin=True):
    """
    Determine alpha beta angles from metrology fiber measurement.

    Parameters
    ------------
    fullTable : pandas.DataFrame
        A merge on holeID of positionerTable and wokCoords DataFrames,
        with additional columns names xyWokMeasMetrology.
    newInvKin : bool
        If True, use new inverse kinematic calculation
        which will never return NaN's

    Returns
    --------
    fullTable : pandas.DataFrame
        the input fullTable with new columns appended
        alphaMeas and betaMeas.
    """

    alpha = []
    beta = []
    for ii, posRow in fullTable.iterrows():

        xWok = float(posRow.xWokMeasMetrology)
        yWok = float(posRow.yWokMeasMetrology)

        xBeta = float(posRow.metX)
        yBeta = float(posRow.metY)

        b = numpy.array([posRow.xWok, posRow.yWok, posRow.zWok])
        iHat = numpy.array([posRow.ix, posRow.iy, posRow.iz])
        jHat = numpy.array([posRow.jx, posRow.jy, posRow.jz])
        kHat = numpy.array([posRow.kx, posRow.ky, posRow.kz])
        la = float(posRow.alphaArmLen)
        alphaOffDeg = float(posRow.alphaOffset)
        betaOffDeg = float(posRow.betaOffset)
        dx = float(posRow.dx)
        dy = float(posRow.dy)

        xBeta = float(posRow.metX)
        yBeta = float(posRow.metY)

        _alpha, _beta = wokToPositioner(
            xWok, yWok, xBeta, yBeta, la,
            alphaOffDeg, betaOffDeg,
            dx, dy, b, iHat, jHat, kHat,
            newInvKin=newInvKin
        )

        alpha.append(_alpha)
        beta.append(_beta)

    fullTable["alphaMeas"] = alpha
    fullTable["betaMeas"] = beta

    return fullTable


# hogg fourier basis fit stuff

IMAX = 32 # maximum integer wave number
DELTAK = 2. * numpy.pi / 10000.0 # wave number spacing in inverse pixels
_wcd = os.environ.get('WOKCALIB_DIR', None)
### BIG WARNING ### if two wok calib dirs are present just take the first one
### should eventually move this to calibrations for better handling
beta_x_apo = None
beta_y_apo = None
beta_x_lco = None
beta_y_lco = None
if _wcd is not None:
    for _path in _wcd.split(":"):
        with open(os.path.join(_path, "beta_x.npy"), "rb") as f:
            beta_x = numpy.load(f)
        with open(os.path.join(_path, "beta_y.npy"), "rb") as f:
            beta_y = numpy.load(f)
        if "/lco/" in _path:
            beta_x_lco = beta_x.copy()
            beta_y_lco = beta_y.copy()
        else:
            beta_x_apo = beta_x.copy()
            beta_y_apo = beta_y.copy()
beta_x = None
beta_y = None


# functions to set up design matrices


def fourier_functions(xs, ys):
    n = len(xs)
    assert len(ys) == n
    fxs = numpy.zeros((n, IMAX * 2 + 2))
    fys = numpy.zeros((n, IMAX * 2 + 2))
    iis = numpy.zeros(IMAX * 2 + 2).astype(int)
    for i in range(IMAX+1):
        fxs[:, i * 2] = numpy.cos(i * DELTAK * xs)
        fys[:, i * 2] = numpy.cos(i * DELTAK * ys)
        iis[i * 2] = i
        fxs[:, i * 2 + 1] = numpy.sin((i + 1) * DELTAK * xs)
        fys[:, i * 2 + 1] = numpy.sin((i + 1) * DELTAK * ys)
        iis[i * 2 + 1] = i + 1
    return fxs, fys, iis


def design_matrix(xs, ys):
    fxs, fys, iis = fourier_functions(xs, ys)
    n, p = fxs.shape
    Xbig = (fxs[:, :, None] * fys[:, None, :]).reshape((n, p * p))
    i2plusj2 = (iis[:, None] ** 2 + iis[None, :] ** 2).reshape(p * p)
    return Xbig[:, i2plusj2 <= IMAX ** 2]


def applyNudgeModel(
        x, y, dxythresh=0.75,
        site="APO", beta_x=None, beta_y=None
    ):

    if beta_x is None and beta_y is None:
        if site == "APO":
            beta_x = beta_x_apo
            beta_y = beta_y_apo
        else:
            beta_x = beta_x_lco
            beta_y = beta_y_lco

    X = design_matrix(x,y)
    dx = X @ beta_x
    dy = X @ beta_y

    # clip dxs and dys at thresh value
    dx[dx>dxythresh] = dxythresh
    dy[dy>dxythresh] = dxythresh

    dx[dx<-dxythresh] = -dxythresh
    dy[dy<-dxythresh] = -dxythresh

    # rejectInds = (numpy.abs(dx) > dxythresh) | (numpy.abs(dy) > dxythresh)

    newX = x - dx
    newY = y - dy

    # newX[rejectInds] = x[rejectInds]
    # newY[rejectInds] = y[rejectInds]

    return newX, newY


class FVCTransformAPO(object):
    # polids = [0, 1, 2, 3, 4, 5, 6, 9, 20, 28, 29]  # Zhao-Burge basis defaults
    polids = numpy.arange(33)
    zbCoeffs = numpy.array([
        3.30790633e-01, 4.13533378e-01, 1.71951213e-03, -9.90141758e-04,
        -4.48460907e-04, -3.68228879e-06, -2.45423918e-06, 3.06249618e-07,
        1.52595326e-06, -1.73037375e-08, 9.09750974e-09, 1.40858150e-08,
        -8.69986544e-11, 2.25009469e-09, 5.94658320e-12, 5.67065198e-12,
        -1.99111356e-12, 6.33175666e-13, -2.82452951e-12, 1.94818210e-12,
        4.03105186e-14, -2.79909483e-14, -1.46249329e-14, -5.04644513e-15,
        1.41549916e-15, 1.27613775e-15, -8.70792948e-15, 2.00843998e-04,
        -1.05007864e-06, 2.48271534e-07, -9.28099433e-10, 6.13172886e-10,
        -1.31799608e-09
    ])
    zbCoeffsFieldCenter = numpy.array([
        -7.25143224e-02,  -3.89886320e-02,   1.21470501e-02,  -4.36545180e-04,
       5.77221153e-05,  -6.80419618e-06,   3.47687303e-06,  -7.22841296e-06,
       3.36246434e-07,  -7.10874816e-07,   9.00745018e-09,   2.40442049e-08,
       1.67481779e-08,   4.90264973e-09,   1.75289911e-11,   2.59249932e-10,
      -8.03748666e-11,   2.04240068e-10,  -9.05440334e-11,   2.60323630e-10,
       1.12381521e-11,  -8.56685856e-14,  -1.37273897e-13,   1.03610076e-14,
      -6.97673027e-13,   1.77283795e-12,   6.96794275e-13,  -5.97070297e-04,
       1.01954593e-06,  -2.25406706e-06,   2.50511002e-08,  -9.82154214e-09,
      -4.41479884e-09
    ])
    telRotAngRef = 135.4
    rotAngDir = 1
    centType = "zbplus2"
    telescopePlateScale = 0.060 # mm/arcsec
    nudgeOffX = 0 #1000 # fix nudge model after FVC resize
    site = "APO"
    centroidMinNpix = 100

    def __init__(
        self,
        fvcImgData,
        positionerCoords,
        telRotAng,
        plotPathPrefix=None,
        positionerTable=calibration.positionerTable,
        wokCoords=calibration.wokCoords,
        fiducialCoords=calibration.fiducialCoords,
        telRotAngRef=None,
        polids=None,
        nudgeAdjust=True,
        zbNormFactor=330
    ):
        """
        Parameters
        -------------
        fvcImgData : numpy.ndarray
            raw image data from the fvc
        positionerCoords : pandas.DataFrame
            DataFrame containing alphaReported and betaReported
            coordinate columns and a positionerID column for each robot
            in the FVC image
        telRotAng : float
            telescope rotator angle in mount coordinate degrees (IPA in sdss
            headers)
        plotPathPrefix : str
            base path for plot output.  Plot routines will append a
            ".<plotname>.pdf" suffix to the prefix supplied.  If None,
            plots will not be generated.
        positionerTable : pandas.DataFrame
            positioner calibration table for robots in the FVC image.
            Default is coordio.defaults.calibrations.positionerTable
        wokCoords : pandas.DataFrame
            wok coordinates for the robots in the FVC image
            Default is coordio.defaults.calibrations.wokCoords
        fiducialCoords : pandas.DataFrame
            coordiates of fiducials in the FVC image
            Default is coordio.defaults.calibrations.fiducialCoords
        telRotAngRef : float
            telescope rotator angle in mount coordinate degrees at
            which xyCCD ~= xyWok directions.
            Default is 135.4 (for APO)
        polids : 1D array or None
            list of integers for selecting zhaoburge basis polynomials.
            Default is supplied by class attribute polids
        nudgeAdjust : boolean
            if True, apply nudge correction to CCD detections
        zbNormFactor : float
            scale wok coordinates by this factor for unit circle
            normalization when fitting ZhaoBurge transform.
        """
        self.nudgeAdjust = nudgeAdjust
        self.zbNormFactor = zbNormFactor

        self.fvcImgData = numpy.array(fvcImgData, dtype=numpy.float32)

        # self.fvcImgData = self.fvcImgData[100:-100,:].copy()

        # apply rough bias/background subtraction
        # rough bias/bg subtract
        imbias = numpy.median(self.fvcImgData, axis=0)
        imbias = numpy.outer(
            numpy.ones(self.fvcImgData.shape[0], dtype=numpy.float32),
            imbias
        )
        im = self.fvcImgData - imbias

        self.bkg = sep.Background(im) #self.fvcImgData)
        bkg_image = self.bkg.back()

        self.data_sub = im - bkg_image

        self.positionerTable = positionerTable.reset_index()
        self.wokCoords = wokCoords.reset_index()
        self.fiducialCoords = fiducialCoords.reset_index()
        self.positionerCoords = positionerCoords.reset_index()
        self.telRotAng = telRotAng
        if telRotAngRef is not None:
            self.telRotAngRef = telRotAngRef
        self.plotPathPrefix = plotPathPrefix
        if polids is not None:
            self.polids = polids

        ft = positionerTable.merge(wokCoords, on="holeID").reset_index()
        ft = ft.merge(positionerCoords, on="positionerID")
        self._fullTable = xyWokFiberFromPositioner(
            ft,
            angleType="Report"
        )

        # construct a matrix for rotating centroids based on telescope
        # rotator angle
        self.ccd2WokRot = (telRotAng - self.telRotAngRef)*self.rotAngDir
        sinRot = numpy.sin(numpy.radians(self.ccd2WokRot))
        cosRot = numpy.cos(numpy.radians(self.ccd2WokRot))
        self.rotMat = numpy.array([
            [cosRot, -sinRot],
            [sinRot, cosRot]
        ])

        self.nCentroid_expect = len(positionerTable) + len(fiducialCoords)
        self.nPositioner_expect = len(positionerTable)
        self.maxCounts = numpy.max(self.fvcImgData)

        ########### populated by self.extractCentroids() #######
        self.centroids = None  # pandas.DataFrame

        self.backgroundSigma = None
        self.winposSigma = None
        self.winposBoxSize = None
        self.ccdRotCenXY = None
        self.nCentroidFound = None

        self.simpleSigma = None
        self.simpleBoxSize = None

        ############## populated by self.fit() ####################
        # pandas.DataFrame join w/ centroids every positioner
        # gets closest match (no clipping), but warnings present
        self.positionerTableMeas = None
        # pandas.DataFrame join w/ centroids
        # gets closest match (no clipping), but warnings present
        self.fiducialCoordsMeas = None
        # panas.DataFrame subset of self.centroids without
        # matches to a robot or fiducial
        self.unmatchedCentroids = None

        self.roughTransform = None
        self.similarityTransform = None
        self.fullTransform = None

        self.maxRoughDist = None
        self.maxMidDist = None
        self.maxFinalDist = None
        self.nOuterFIF_expect = None
        self.nOuterFIF_found = None
        self.nFIF_expect = None
        self.nFIF_found = None
        self.fiducialRMS = None
        self.nFiducialWarn = None
        self.positionerRMS = None
        self.positionerRMS_clipped = None
        self.nPositionerWarn = None
        self.positionerWarnList = None

    def getMetadata(self):
        """Get a list of data that can be easily stuffed in a fits
        header.
        """

        metaDataList = [
            ("FVC_NWRN", self.nPositionerWarn, "number of robots out of measurement spec"),
            ("FVC_MAXD", self.maxFinalDist, "distance beyond to consider robot out of spec (mm)"),
            ("FVC_CNTT", self.centType, "centroid type used for fitting"),
            ("FVC_BSIG", self.backgroundSigma, "above background sigma for centroid detection"),
            ("FVC_MNPX", self.centroidMinNpix, "minimum number of pixels for a valid centroid"),
            ("FVC_WSIG", self.winposSigma, "sigma for winpos centroid algorithm"),
            ("FVC_WBSZ", self.winposBoxSize, "box size for winpos centroid algorithm (pix)"),
            ("FVC_SSIG", self.simpleSigma, "sigma for simple centroid algorithm (pix)"),
            ("FVC_SPLM", self.simpleBoxSize, "box size for simple centroid algorithm (pix)"),
            ("FVC_RMS", self.positionerRMS, "robot rms (mm)"),
            ("FVC_FRMS", self.fiducialRMS, "fiducial rms (mm)"),
            ("FVC_CRMS", self.positionerRMS_clipped, "in-spec (outlier-clipped) robot rms (mm)"),
            ("FVC_SCL", self.fullTransform.simTrans.scale, "FVC model fit scale"),
            ("FVC_TRAX", self.fullTransform.simTrans.translation[0], "FVC model fit X translation"),
            ("FVC_TRAY", self.fullTransform.simTrans.translation[1], "FVC model fit Y translation"),
            ("FVC_ROT", numpy.degrees(self.fullTransform.simTrans.rotation), "FVC model fit rotation (deg)"),
            ("FVC_ZNRM", self.zbNormFactor, "ZB normilization factor")
        ]

        # add in ZB coeffs
        for polid, coeff in zip(self.polids, self.fullTransform.coeffs):
            metaDataList.append(
                ("FVC_ZB%i"%polid, coeff, "zhao-burge transform coeff for polid %i"%polid)
            )

        return metaDataList

    def extractCentroids(
        self,
        centroidMinNpix=None,
        backgroundSigma=3.5,
        winposSigma=0.7,
        winposBoxSize=3,
        simpleSigma=5,
        simpleBoxSize=19,
        beta_x=None,
        beta_y=None
    ):
        """
        Find centroids in the fvc image, stores result in
        self.centroids attribute (a pandas.DataFrame)

        Parameters
        -----------
        centroidMinNpix : int
            minimum number of pixels that belong to a bona fide detection
        backgoundSigma : float
            sigma above background that belong to a bona fide detection
        winposSigma : float
            used by sep.winpos.  Gaussian sigma used for weighting pixels.
            Pixels within a circular aperture of radius 4*winposSigma are
            included.
        winposBoxSize : int
            odd integer.  winpos centroids will be searched for in a
            winposBoxSize x winposBoxSize centered on the peak
            pixel.
        """
        if centroidMinNpix is not None:
            self.centroidMinNpix = centroidMinNpix
        self.backgroundSigma = backgroundSigma
        self.winposSigma = winposSigma
        self.winposBoxSize = winposBoxSize
        self.ccdRotCenXY = numpy.array(self.data_sub.shape) / 2.
        self.simpleSigma = simpleSigma
        self.simpleBoxSize = simpleBoxSize

        if winposBoxSize % 2 == 0 or winposBoxSize <= 0:
            raise CoordIOError("winposBoxSize must be a positive odd integer")

        # ccdRotCenXY = numpy.array(ccdRotCenXY).squeeze()
        # if len(ccdRotCenXY) != 2:
        #     raise CoordIOError("ccdRotCenXY must be a 2 element vector")


        objects = sep.extract(
            self.data_sub,
            backgroundSigma,
            err=self.bkg.globalrms,
        )

        # get rid of obvious bogus detections
        objects = objects[objects["npix"] > self.centroidMinNpix]
        # remove detections near edges of chip
        # (causes issues for unlucky winpos detections)
        # objects = objects[objects["x"] > 500]
        # objects = objects[objects["x"] < 7500]
        # objects = objects[objects["y"] > 30]
        # objects = objects[objects["y"] < 5970]

        # update based on CCD distortion model the "true location"
        # of the fibers
        if self.data_sub.shape[1] < 8000 and self.nudgeAdjust:
            # ccd was trimmed
            xNudge, yNudge = applyNudgeModel(objects["x"]+self.nudgeOffX, objects["y"], site=self.site, beta_x=beta_x, beta_y=beta_y)
            xNudge = xNudge - self.nudgeOffX
        else:
            xNudge, yNudge = applyNudgeModel(objects["x"], objects["y"], site=self.site, beta_x=beta_x, beta_y=beta_y)


        # don't fit anything with an absolute correction > 0.75 pixels
        # rejectInds = (numpy.abs(xNudge) > 0.75) | (numpy.abs(yNudge) > 0.75)


        # create mask and re-extract using winpos algorithm
        # maskArr = numpy.ones(self.data_sub.shape, dtype=bool)
        # boxRad = numpy.floor(winposBoxSize/2)
        # boxSteps = numpy.arange(-boxRad, boxRad+1, dtype=int)

        # for ii in range(len(objects)):
        #     _xm = objects["xcpeak"][ii]
        #     _ym = objects["ycpeak"][ii]
        #     for xstep in boxSteps:
        #         for ystep in boxSteps:
        #             maskArr[_ym + ystep, _xm + xstep] = False

        # xNew, yNew, flags = sep.winpos(
        #     self.data_sub,
        #     objects["xcpeak"],
        #     objects["ycpeak"],
        #     sig=winposSigma,
        #     mask=maskArr
        # )

        # refinexy requires a square array
        # trim and adjust results accordingly

        # find the longest dimension
        clipDim = numpy.argmax(self.data_sub.shape)

        if clipDim == 0:
            yOff = int((self.data_sub.shape[0]-self.data_sub.shape[1])/2)
            im = self.data_sub[yOff:-yOff, :].copy()

            xSimple, ySimple = refinexy(
                im, objects["x"], objects["y"]-yOff,
                psf_sigma=self.simpleSigma, cutout=self.simpleBoxSize
            )
            ySimple = ySimple + yOff
        else:
            xOff = int((self.data_sub.shape[1]-self.data_sub.shape[0])/2)
            im = self.data_sub[:, xOff:-xOff].copy()

            xSimple, ySimple = refinexy(
                im, objects["x"]-xOff, objects["y"],
                psf_sigma=self.simpleSigma, cutout=self.simpleBoxSize
            )
            xSimple = xSimple + xOff

        objects = pandas.DataFrame(objects)

        # objects["xWinpos"] = xNew
        # objects["yWinpos"] = yNew
        objects["xSimple"] = xSimple
        objects["ySimple"] = ySimple
        objects["xNudge"] = xNudge
        objects["yNudge"] = yNudge

        # rotate raw centroids by rotator angle
        _centTypes = [
            ["x", "y"],
            # ["xWinpos", "yWinpos"],
            ["xSimple", "ySimple"],
            ["xNudge", "yNudge"]
        ]

        for _centType in _centTypes:
            xy = objects[_centType].to_numpy()
            xyRot = (self.rotMat @ (xy - self.ccdRotCenXY).T).T + self.ccdRotCenXY
            objects[_centType[0]+"Rot"] = xyRot[:,0]
            objects[_centType[1]+"Rot"] = xyRot[:,1]

        objects["centroidID"] = list(range(len(objects)))

        self.centroids = objects
        self.nCentroids = len(objects)

        return self.centroids

    def fit(
        self,
        centType=None,
        maxRoughDist=10,
        maxMidDist=4,
        maxFinalDist=0.5,
        newInvKin=True
    ):
        """
        Calculate xy wok positions of centroids.  Store results
        in self.fullTable (a pandas.DataFrame)

        Parameters
        -----------
        centType : str
            one of "zbplus2", zbplus", "zbminus", "sep", or "nudge", default is "nudge"
        maxRoughDist : float
            Max distance for an outer fiducial match (rough mm)
        maxMidDist : float
            Max distance for all fiducial matches after similarity
            transform based on outer fiducials (~ mm)
        maxFinalDist : float
            Max distance for positioner and fiducial matches
            after the full similarity + ZB transform (mm)
        newInvKin : bool
            If True use new inverse kinematics
        """
        if centType is not None:
            centType = centType.lower()
            self.centType = centType
        self.maxRoughDist = maxRoughDist
        self.maxMidDist = maxMidDist
        self.maxFinalDist = maxFinalDist
        self.newInvKin = newInvKin

        if self.centroids is None:
            raise CoordIOError("Must run extractCentroids before fit")
        if self.centType not in ["zbplus2", "zbplus", "zbminus", "sep", "nudge", "simple"]:
            raise CoordIOError("unknown centType: %s"%str(self.centType))

        xyMetFiber = self._fullTable[
            ["xWokReportMetrology", "yWokReportMetrology"]
        ].to_numpy()

        xyWokFIF = self.fiducialCoords[["xWok", "yWok"]].to_numpy()

        # if self.centType == "winpos":
        #     xyCCDRot = self.centroids[["xWinposRot", "yWinposRot"]].to_numpy()
        #     xyCCD = self.centroids[["xWinpos", "yWinpos"]].to_numpy()
        if self.centType == "sep":
            xyCCDRot = self.centroids[["xRot", "yRot"]].to_numpy()
            xyCCD = self.centroids[["x", "y"]].to_numpy()
        elif self.centType == "simple":
            xyCCDRot = self.centroids[["xSimpleRot", "ySimpleRot"]].to_numpy()
            xyCCD = self.centroids[["xSimple", "ySimple"]].to_numpy()
        elif self.centType in ["nudge", "zbplus", "zbminus", "zbplus2"]:
            xyCCDRot = self.centroids[["xNudgeRot", "yNudgeRot"]].to_numpy()
            xyCCD = self.centroids[["xNudge", "yNudge"]].to_numpy()
        else:
            raise CoordIOError("unknown centroid type passed to fit")

        # first do a rough transform
        self.roughTransform = RoughTransform(
            xyCCDRot,
            numpy.vstack((xyMetFiber, xyWokFIF))
        )
        xyWokRough = self.roughTransform.apply(xyCCDRot)

        # just grab outer fiducials for first pass, they're easier to identify
        # centroids lying at radii > 310 must be outer ring fiducials
        rWokFIF = numpy.linalg.norm(xyWokFIF, axis=1)
        xyWokOuterFIF = xyWokFIF[rWokFIF > 310]

        # associate the centroids to the outer wok FIDs
        xyWokOuterFIF_idx, xyWokRough_idx, distances = arg_nearest_neighbor(
            xyWokOuterFIF,
            xyWokRough,
            maxRoughDist,
        )


        self.nOuterFIF_expect = len(xyWokOuterFIF)
        xyWokOuterFIF = xyWokOuterFIF[xyWokOuterFIF_idx]
        self.nOuterFIF_found = len(xyWokOuterFIF)


        # plot the rough transform
        if self.plotPathPrefix is not None:
            assoc_found = [xyWokOuterFIF, xyWokRough[xyWokRough_idx]]
            plotFVCResults(
                self.plotPathPrefix + ".roughTransform.pdf",
                xyFitCentroids=xyWokRough,
                xyMetFiber=xyMetFiber,
                xyFIF=xyWokFIF,
                assoc_used=assoc_found,
                title="Rough Transform and Outer Fiducial Associations\n" + \
                        "missing %i fiducials"%(
                            self.nOuterFIF_expect - self.nOuterFIF_found
                        )
            )

        # use associations from rough transform to fit a full
        # transform using only outer fiducial ring
        # pass zb = false to just fit translation rotation and scale

        xyCCDOuterFIF = xyCCD[xyWokRough_idx]

        self.similarityTransform = SimilarityTransform()
        self.similarityTransform.estimate(xyCCDOuterFIF, xyWokOuterFIF)
        xyWokMeas = self.similarityTransform(xyCCD)

        # associate centroids with all fiducials in grid, now that we're closer

        xyWokFIF_idx, xyWokMeas_idx, distances = arg_nearest_neighbor(
            xyWokFIF,
            xyWokMeas,
            maxMidDist,
        )

        # find unmatched fiducials
        unmatchedFIF_idx = numpy.array(list(set(range(len(xyWokFIF))) - set(xyWokFIF_idx)), dtype=int)
        xyFIFLost = xyWokFIF[unmatchedFIF_idx]

        self.nFIF_expect = len(xyWokFIF)
        _xyWokFIF = xyWokFIF[xyWokFIF_idx]
        self.nFIF_found = len(_xyWokFIF)

        #plot similarity transform
        if self.plotPathPrefix is not None:
            assoc_found = [_xyWokFIF, xyWokMeas[xyWokMeas_idx]]
            plotFVCResults(
                self.plotPathPrefix + ".similarityTransform.pdf",
                xyFitCentroids=xyWokMeas,
                xyMetFiber=xyMetFiber,
                xyFIF=xyWokFIF,
                assoc_used=assoc_found,
                title="Similarity Transform and Full Fiducial Associations\n" + \
                        "missing %i fiducials"%(
                            self.nFIF_expect - self.nFIF_found
                        ),
                xyFIFLost=xyFIFLost
            )

        # finally, do the full ZB transform based on all found FIF locations
        xyCCDFIF = xyCCD[xyWokMeas_idx]
        self.fullTransform = ZhaoBurgeTransform(
            xyCCDFIF,
            _xyWokFIF,
            polids=self.polids,
            normFactor=self.zbNormFactor
        )

        xyWokMeas = self.fullTransform.apply(xyCCD)

        # update the final measurements for positioners
        # and fiducials based on the fullTransform
        # dont enforce any minimum distance
        xyWokFIF_idx, xyWokMeas_idx, distances = arg_nearest_neighbor(
            xyWokFIF,
            xyWokMeas,
            atol=None,  # no min distance
        )

        fiducialMeas = self.fiducialCoords.iloc[xyWokFIF_idx].reset_index()
        centroidMatched = self.centroids.iloc[xyWokMeas_idx].reset_index()

        fiducialMeas = pandas.concat(
            [fiducialMeas, centroidMatched], axis=1
        )

        fiducialMeas["xWokMeas"] = xyWokMeas[xyWokMeas_idx, 0]
        fiducialMeas["yWokMeas"] = xyWokMeas[xyWokMeas_idx, 1]
        # add distances in match
        fiducialMeas["wokErr"] = distances
        fiducialMeas["wokErrWarn"] = distances > maxFinalDist

        self.fiducialCoordsMeas = fiducialMeas

        dx = self.fiducialCoordsMeas.xWokMeas - self.fiducialCoordsMeas.xWok
        dy = self.fiducialCoordsMeas.yWokMeas - self.fiducialCoordsMeas.yWok
        self.fiducialRMS = numpy.sqrt(numpy.mean(dx**2 + dy**2))
        self.nFiducialWarn = sum(fiducialMeas.wokErrWarn)

        xyWokMet_idx, xyWokMeas_idx, distances = arg_nearest_neighbor(
            xyMetFiber,
            xyWokMeas,
            atol=None, # no min distance
        )

        positionerMeas = self._fullTable.iloc[xyWokMet_idx].reset_index()
        centroidMatched = self.centroids.iloc[xyWokMeas_idx].reset_index()
        positionerMeas = pandas.concat(
            [positionerMeas, centroidMatched], axis=1
        )

        positionerMeas["xWokMeasMetrology"] = xyWokMeas[xyWokMeas_idx, 0]
        positionerMeas["yWokMeasMetrology"] = xyWokMeas[xyWokMeas_idx, 1]
        positionerMeas["wokErr"] = distances
        positionerMeas["wokErrWarn"] = distances > maxFinalDist

        #### apply zb corrections if centype is zbplus or zbminus
        if self.centType.startswith("zb") and self.zbCoeffs is not None:
            _xWok = positionerMeas.xWokMeasMetrology.to_numpy()
            _yWok = positionerMeas.yWokMeasMetrology.to_numpy()
            dx_zb_fit, dy_zb_fit = getZhaoBurgeXY(
                self.polids, self.zbCoeffs, _xWok, _yWok
            )
            # multiply by scale (arcsec to mm)

            if self.centType == "zbminus":
                dx_zb_fit *= -1
                dy_zb_fit *= -1
            elif self.centType == "zbplus2":
                # fit the inner region rWok < 100 mm with another
                # set of zb coeffs determined from dither analysis
                _xWok = _xWok + dx_zb_fit
                _yWok = _yWok + dy_zb_fit
                keep = numpy.sqrt(_xWok**2+_yWok**2) < 100
                dx_zb_fit2, dy_zb_fit2 = getZhaoBurgeXY(
                    self.polids, self.zbCoeffsFieldCenter,
                    _xWok[keep], _yWok[keep]
                )
                # dx2 = numpy.zeros(len(_xWok))
                # dy2 = numpy.zeros(len(_xWok))
                # dx2[keep] = dx_zb_fit2
                # dy2[keep] = dy_zb_fit2
                dx_zb_fit[keep] += dx_zb_fit2
                dy_zb_fit[keep] += dy_zb_fit2

            dx_zb_fit *= self.telescopePlateScale
            dy_zb_fit *= self.telescopePlateScale

        else:
            dx_zb_fit = numpy.zeros(len(positionerMeas))
            dy_zb_fit = numpy.zeros(len(positionerMeas))

        positionerMeas["xWokMeasMetrology"] = positionerMeas["xWokMeasMetrology"] + dx_zb_fit
        positionerMeas["yWokMeasMetrology"] = positionerMeas["yWokMeasMetrology"] + dy_zb_fit
        positionerMeas["xWokAdjMetrology"] = dx_zb_fit
        positionerMeas["yWokAdjMetrology"] = dy_zb_fit


        self.positionerTableMeas = positionerMeas

        dx = self.positionerTableMeas.xWokMeasMetrology - \
             self.positionerTableMeas.xWokReportMetrology
        dy = self.positionerTableMeas.yWokMeasMetrology - \
             self.positionerTableMeas.yWokReportMetrology
        sqerr = dx**2 + dy**2

        self.positionerRMS = numpy.sqrt(numpy.mean(sqerr))
        self.positionerRMS_clipped = numpy.sqrt(
            numpy.mean(sqerr[distances < maxFinalDist])
        )
        self.nPositionerWarn = numpy.sum(distances > maxFinalDist)
        _pwl = positionerMeas[positionerMeas.wokErrWarn]["positionerID"].to_numpy()
        self.positionerWarnList = _pwl


        # find unmatched remaining centroids
        centroidsUsed = list(set(
            list(self.positionerTableMeas.centroidID) + \
            list(self.fiducialCoordsMeas.centroidID)
        ))

        unusedCentroids_idx = ~self.centroids["centroidID"].isin(centroidsUsed)
        self.unmatchedCentroids = self.centroids[unusedCentroids_idx]
        unusedCentroidsXY = xyWokMeas[unusedCentroids_idx]


        xyMetWarn = self.positionerTableMeas[self.positionerTableMeas.wokErrWarn]
        xyMetWarn = xyMetWarn[
            ["xWokReportMetrology", "yWokReportMetrology"]
        ].to_numpy()



        if self.plotPathPrefix is not None:
            pwarnStr = ",".join("%i"%pid for pid in self.positionerWarnList)
            title = "Zhao-Burge Transform of all Detections\n"
            title += "positioner warnings: %s"%pwarnStr

            # plot the full transform
            plotFVCResults(
                self.plotPathPrefix + ".zhaoBurgeTransform.pdf",
                xyFitCentroids=xyWokMeas,
                xyMetFiber=xyMetFiber,
                xyFIF=xyWokFIF,
                positionerIDs=list(self._fullTable.positionerID),
                xyFitCentroidsUnmatched=unusedCentroidsXY,
                xyFiberWarn=xyMetWarn,
                title=title,
                xyFIFLost=xyFIFLost
            )

        # lastly compute measured alpha/beta and measured boss/Ap
        # locations of fibers based on metrology fiber location

        self.positionerTableMeas = alphaBetaFromMetMeas(
            self.positionerTableMeas, newInvKin=newInvKin
        )

        self.positionerTableMeas = xyWokFiberFromPositioner(
            self.positionerTableMeas, angleType="Meas", doMetrology=False
        )

        if self.centType in ["nudge", "zbplus", "zbminus", "zbplus2"]:
            self.positionerTableMeas["xFVC"] = self.positionerTableMeas["xNudge"]
            self.positionerTableMeas["yFVC"] = self.positionerTableMeas["yNudge"]
            self.fiducialCoordsMeas["xFVC"] = self.fiducialCoordsMeas["xNudge"]
            self.fiducialCoordsMeas["yFVC"] = self.fiducialCoordsMeas["yNudge"]
        # elif self.centType == "winpos":
        #     self.positionerTableMeas["xFVC"] = self.positionerTableMeas["xWinpos"]
        #     self.positionerTableMeas["yFVC"] = self.positionerTableMeas["yWinpos"]
        #     self.fiducialCoordsMeas["xFVC"] = self.fiducialCoordsMeas["xWinpos"]
        #     self.fiducialCoordsMeas["yFVC"] = self.fiducialCoordsMeas["yWinpos"]
        elif self.centType == "simple":
            self.positionerTableMeas["xFVC"] = self.positionerTableMeas["xSimple"]
            self.positionerTableMeas["yFVC"] = self.positionerTableMeas["ySimple"]
            self.fiducialCoordsMeas["xFVC"] = self.fiducialCoordsMeas["xSimple"]
            self.fiducialCoordsMeas["yFVC"] = self.fiducialCoordsMeas["ySimple"]
        elif self.centType == "sep":
            self.positionerTableMeas["xFVC"] = self.positionerTableMeas["x"]
            self.positionerTableMeas["yFVC"] = self.positionerTableMeas["y"]
            self.fiducialCoordsMeas["xFVC"] = self.fiducialCoordsMeas["x"]
            self.fiducialCoordsMeas["yFVC"] = self.fiducialCoordsMeas["y"]


class FVCTransformLCO(FVCTransformAPO):
    # polids = [0, 1, 2, 3, 4, 5, 6, 9, 20, 28, 29]  # Zhao-Burge basis defaults, best so far

    polids = numpy.array([0, 1, 2, 3, 4, 5, 6, 9, 20, 27, 28, 29, 30])
    zbCoeffs = None
    zbCoeffsFieldCenter = None
    telRotAngRef = 89  # rotator angle that puts xyWok aligned with xyCCD on FVC image
    rotAngDir = -1
    centType = "nudge"
    telescopePlateScale = 0.092 # mm/arcsec
    nudgeOffX = 0  # fix nudge model after FVC resize
    site = "LCO"
    centroidMinNpix = 20



import numpy
import pytest
from numpy.testing import assert_array_almost_equal

from coordio import CoordIOError, Guide, Site, Tangent, calibration, defaults


numpy.random.seed(0)


CCD_SIZE_MM = defaults.GFA_CHIP_CENTER * 2 * defaults.GFA_PIXEL_SIZE / \
           defaults.MICRONS_PER_MM

nCoords = 100
apoSite = Site("APO")
lcoSite = Site("LCO")

# # set time (not used but required for Observed coords)
apoSite.set_time(2458863, scale='TAI')
lcoSite.set_time(2458863, scale='TAI')


def test_guide_cycle():
    tangX = numpy.random.uniform(
        -CCD_SIZE_MM/2, CCD_SIZE_MM/2, size=nCoords
    )
    tangY = numpy.random.uniform(
        -CCD_SIZE_MM/2, CCD_SIZE_MM/2, size=nCoords
    )
    tangZ = numpy.random.uniform(-0.1, 0.1, size=nCoords)

    tangXYZ = numpy.array([tangX, tangY, tangZ]).T

    for site in [apoSite, lcoSite]:
        for holeID in calibration.VALID_GUIDE_IDS:
            scaleFactor = numpy.random.uniform(.99, 1.01)
            binX = numpy.random.choice([1,2])
            binY = numpy.random.choice([1,2])
            obsAngle = numpy.random.uniform(0,360)
            tc = Tangent(
                tangXYZ, site=site, holeID=holeID,
                scaleFactor=scaleFactor, obsAngle=obsAngle
            )
            gc = Guide(tc, binX=binX, binY=binY)

            _tc = Tangent(
                gc, site=site, holeID=holeID,
                scaleFactor=scaleFactor, obsAngle=obsAngle
            )

            assert_array_almost_equal(tc.xProj, _tc.xProj, decimal=10)
            assert_array_almost_equal(tc.xProj, _tc[:,0], decimal=10)
            assert_array_almost_equal(tc.yProj, _tc.yProj, decimal=10)
            assert_array_almost_equal(tc.yProj, _tc[:,1], decimal=10)
            assert_array_almost_equal(_tc[:,2], numpy.zeros(nCoords), decimal=10)
            assert True not in gc.guide_warn


def test_guide_warn():

    tangX = numpy.random.uniform(-CCD_SIZE_MM/2, CCD_SIZE_MM/2, size=nCoords)
    tangY = numpy.random.uniform(-CCD_SIZE_MM/2, CCD_SIZE_MM/2, size=nCoords)
    tangZ = numpy.random.uniform(-0.1, 0.1, size=nCoords)

    badInds = [2, 3, 4, 5]
    # plug in values that will land outside the chip
    tangX[2] = -5*CCD_SIZE_MM

    tangY[3] = 2*CCD_SIZE_MM

    tangX[4] = CCD_SIZE_MM
    tangY[4] = -CCD_SIZE_MM

    tangY[5] = -5*CCD_SIZE_MM

    tangXYZ = numpy.array([tangX, tangY, tangZ]).T

    for site in [apoSite, lcoSite]:
        for holeID in calibration.VALID_GUIDE_IDS:
            scaleFactor = numpy.random.uniform(.99, 1.01)
            binX = numpy.random.choice([1,2])
            binY = numpy.random.choice([1,2])
            obsAngle = numpy.random.uniform(0,360)
            tc = Tangent(
                tangXYZ, site=site, holeID=holeID,
                scaleFactor=scaleFactor, obsAngle=obsAngle
            )
            gc = Guide(tc, xBin=binX, yBin=binY)

            for ii, warn in enumerate(gc.guide_warn):
                if ii in badInds:
                    assert warn == True
                else:
                    assert warn == False


@pytest.mark.xfail(reason="GFAs not yet handled.")
def test_guide_fail():
    wl = defaults.INST_TO_WAVE["Boss"]

    tangX = numpy.random.uniform(0, CCD_SIZE_MM, size=nCoords)
    tangY = numpy.random.uniform(0, CCD_SIZE_MM, size=nCoords)
    tangZ = numpy.random.uniform(-0.1, 0.1, size=nCoords)
    tangXYZ = numpy.array([tangX, tangY, tangZ]).T

    tc = Tangent(
        tangXYZ, site=apoSite, holeID="GFA-S1", wavelength=wl
    )

    with pytest.raises(CoordIOError) as ee:
        # bad wavelength supplied
        Guide(tc)
    assert "non-guide wavelength" in str(ee)

    tc = Tangent(
        tangXYZ, site=apoSite, holeID="R0C1", wavelength=wl
    )

    with pytest.raises(CoordIOError) as ee:
        # bad hole supplied
        Guide(tc)
    assert "wok hole" in str(ee)

    gx = numpy.random.uniform(50, 80, size=nCoords)
    gy = numpy.random.uniform(50, 80, size=nCoords)
    gxy = numpy.array([gx,gy]).T

    gc = Guide(gxy)

    with pytest.raises(CoordIOError) as ee:
        Tangent(gc, site=apoSite, holeID="R0C1")
    assert "non-GFA location" in str(ee)

    with pytest.raises(CoordIOError) as ee:
        Tangent(gc, site=apoSite, holeID="GFA-S1", wavelength=wl)
    assert "non-guide wavelength" in str(ee)



import numpy
import pytest

from coordio import Field, FocalPlane, Observed, Site, Tangent, Wok
from coordio.defaults import APO_MAX_FIELD_R, LCO_MAX_FIELD_R, calibration


@pytest.fixture
def prepare_test():
    numpy.random.seed(0)

    # from sdssconv.fieldCoords import parallacticAngle
    apoSite = Site("APO")
    lcoSite = Site("LCO")

    # set time (not used but required for Observed coords)
    apoSite.set_time(2458863, scale='TAI')
    lcoSite.set_time(2458863, scale='TAI')

    # start with some field coords and propagate them
    nCoords = 2000
    fcAPO = Observed([[80, 120]], site=apoSite)
    fcLCO = Observed([[80, 120]], site=lcoSite)
    thetaField = numpy.random.uniform(0, 360, size=nCoords)
    phiFieldLCO = numpy.random.uniform(0, LCO_MAX_FIELD_R, size=nCoords)
    phiFieldAPO = numpy.random.uniform(0, APO_MAX_FIELD_R, size=nCoords)

    fieldAPO = Field(
        numpy.array([thetaField, phiFieldAPO]).T,
        field_center=fcAPO
    )

    fieldLCO = Field(
        numpy.array([thetaField, phiFieldLCO]).T,
        field_center=fcLCO
    )

    focalAPO = FocalPlane(fieldAPO, site=apoSite)
    focalLCO = FocalPlane(fieldLCO, site=lcoSite)

    yield (apoSite, lcoSite, focalAPO, focalLCO)


def test_focal_wok_cycle(prepare_test):
    apoSite, lcoSite, focalAPO, focalLCO = prepare_test

    for site, focalCoords in zip([lcoSite, apoSite], [focalLCO, focalAPO]):
        obsAngle = numpy.random.uniform(0, 360)
        wokCoords = Wok(focalCoords, site=site, obsAngle=obsAngle)
        _focalCoords = FocalPlane(wokCoords, site=site)
        numpy.testing.assert_array_almost_equal(
            focalCoords, _focalCoords, decimal=10
        )


def test_wok_tangent_cycle(prepare_test):
    apoSite, lcoSite, focalAPO, focalLCO = prepare_test

    for site, focalCoords in zip([lcoSite, apoSite], [focalLCO, focalAPO]):
        obsAngle = numpy.random.uniform(0, 360)
        wokCoords = Wok(focalCoords, site=site, obsAngle=obsAngle)
        site_hole_ids = calibration.positionerTable.loc[site.name].index.tolist()
        for holeID in calibration.VALID_HOLE_IDS:
            if holeID not in site_hole_ids:
                continue
            scaleFactor = numpy.random.uniform(.97, 1.03)
            tangentCoords = Tangent(
                wokCoords, site=site, holeID=holeID,
                scaleFactor=scaleFactor, obsAngle=obsAngle
            )
            _wokCoords = Wok(tangentCoords, site=site, obsAngle=obsAngle)
            numpy.testing.assert_array_almost_equal(
                wokCoords, _wokCoords, decimal=10
            )



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy

from . import calibration, conv, defaults
from .coordinate import Coordinate, Coordinate2D, verifySite
from .exceptions import CoordIOError


if TYPE_CHECKING:
    from .site import Site


class PositionerBase(Coordinate2D):
    """A representation of Positioner coordinates.  Alpha/Beta coordinates
    in degrees.

    When converting from tangent coordinates,
    Robot's are always "right handed" with alpha in [0, 360], beta in [0, 180]

    Parameters
    ------------
    value : numpy.ndarray
        A Nx2 array with columns [alpha, beta] in degrees
    site : `.Site`
        mandatory parameter
    holeID : str
        valid identifier, one of calibration.VALID_HOLE_IDS

    """

    __extra_params__ = ["site", "holeID"]
    __warn_arrays__ = ["positioner_warn"]
    __fiber_type__ = None  # overridden by subclasses

    # For typing.
    positioner_warn: numpy.ndarray
    holeID: str
    site: Site

    def __new__(cls, value, **kwargs):

        verifySite(kwargs, strict=False)

        holeID = kwargs.get("holeID", None)
        if holeID is None:
            raise CoordIOError("Must specify holeID for Positioner Coords")
        if holeID not in calibration.VALID_HOLE_IDS:
            raise CoordIOError("Must be valid holeID for Positioner Coords")

        if isinstance(value, Coordinate):
            if value.coordSysName == "Tangent":
                if holeID.startswith("GFA"):
                    raise CoordIOError(
                        "Guide holeID supplied for Positioner coord"
                    )
                # going from 3D to 2D coordsys
                # initialize array
                initArr = numpy.zeros((len(value), 2))
                obj = super().__new__(cls, initArr, **kwargs)
                obj._loadFiberData()
                obj._fromTangent(value)
            else:
                raise CoordIOError(
                    'Cannot convert to Positioner from %s' % value.coordSysName
                )
        else:
            obj = super().__new__(cls, value, **kwargs)
            obj._loadFiberData()
            obj._fromRaw()

        if obj.__fiber_type__ not in ["Apogee", "Boss", "Metrology"]:
            raise CoordIOError("valid __fiber_type__ must be specified!")

        return obj

    def _loadFiberData(self):
        fiberData = defaults.getPositionerData(self.site.name, self.holeID)

        if self.__fiber_type__ == "Metrology":
            xFiber = fiberData[1]
            yFiber = fiberData[2]

        elif self.__fiber_type__ == "Apogee":
            xFiber = fiberData[3]
            yFiber = fiberData[4]

        elif self.__fiber_type__ == "Boss":
            xFiber = fiberData[5]
            yFiber = fiberData[6]
        else:
            raise CoordIOError("Fiber not specified for positioner coords")

        self.alphaArmLength = fiberData[0]
        self.xFiber = xFiber
        self.yFiber = yFiber

        self.alphaOffset = fiberData[7]
        self.betaOffset = fiberData[8]

    def _fromTangent(self, tangentCoords):
        """Convert from tangent coords to alpha betas
        """
        xTangent = tangentCoords.xProj
        yTangent = tangentCoords.yProj

        # this will always return a right hand orientation
        alphaDeg, betaDeg, isOK = conv.tangentToPositioner(
            xTangent,
            yTangent,
            self.xFiber,
            self.yFiber,
            self.alphaArmLength,
            self.alphaOffset,
            self.betaOffset
        )

        self[:, 0] = alphaDeg
        self[:, 1] = betaDeg
        self.positioner_warn[:] = isOK == False
        self._fromRaw()

    def _fromRaw(self):

        # TODO: maybe not do this?
        self = self % 360


class PositionerBoss(PositionerBase):
    __fiber_type__ = "Boss"


class PositionerApogee(PositionerBase):
    __fiber_type__ = "Apogee"


class PositionerMetrology(PositionerBase):
    __fiber_type__ = "Metrology"



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

import numpy

from . import calibration, conv, defaults
from .coordinate import Coordinate, Coordinate3D, verifySite, verifyWavelength
from .exceptions import CoordIOError, CoordIOUserWarning
from .telescope import FocalPlane
from .wok import Wok


if TYPE_CHECKING:
    from .site import Site


def _getRayOrigins(site, holeID, scaleFactor, obsAngle):
    """Return the location of the spherical focal surface center
    in tangent coordinates for a specific wok location

    site : `.Site
        site with name attribute "APO" or "LCO"
    holeID : str
        A valid hole identifier
    scaleFactor : float
        Scale factor for wok expansion
    obsAngle : float
        observation (position) angle in Deg

    Returns
    --------
    apCen : numpy.ndarray
        3 element array in focal plane coords for center of apogee focal sphere
    bossCen : numpy.ndarray
        3 element array in focal plane coords for center of boss focal sphere
    gfaCen : numpy.ndarray
        3 element array in focal plane coords for center of gfa focal sphere

    """
    outList = []
    direction = "focal"  # irrelevant, just getting sphere param
    for waveCat in ["Apogee", "Boss", "GFA"]:
        R, b, c0, c1, c2, c3, c4 = defaults.getFPModelParams(site.name,
                                                             direction,
                                                             waveCat)
        fpXYZ = [[0, 0, b]]  # sphere's center in focal plane coords
        # JSG for Conor: is this right?
        wavelength = defaults.INST_TO_WAVE['GFA']
        fpCoords = FocalPlane(fpXYZ, site=site, wavelength=wavelength)
        wokCoords = Wok(fpCoords, site=site, obsAngle=obsAngle)
        tanCoords = TangentNoProj(
            wokCoords,
            site=site,
            holeID=holeID,
            scaleFactor=scaleFactor,
            obsAngle=obsAngle,
            wavelength=wavelength
        )
        tanCoords = numpy.array(tanCoords).squeeze()
        outList.append(tanCoords)

    return tuple(outList)


class Tangent(Coordinate3D):
    """A representation of Tangent Coordinates.  A 3D Cartesian coordinate
    system.  The xy plane of this coordinate system is tangent to the wok
    surface at a specific location (one of holeID in etc/wokCoords.csv).
    For positioners, +x axis is a aligned with the alpha=0 direction.  For
    GFA's the -y axis points toward the wok coordinate system origin. +z axis
    (generally) points from the earth to sky.  Origin lies 143 mm above the
    wok surface.  This puts the Tangent xy plane in the plane of the fiber or
    chip.

    Parameters
    ------------
    value : numpy.ndarray
        A Nx3 array where [x,y,z] are columns. Or `.Wok`.  Or `.Positioner`.
        Or `.Guide`.
    site : `.Site`
        Site name determines which wok parameters to use.  Mandatory parameter.
    holeID : str
        Calid identifier(s), one of calibration.VALID_HOLE_IDS
    scaleFactor : float
        Multiplicative factor to apply, modeling thermal expansion/contraction
        of wok holes with respect to each other.  Defaults to 1
    wavelength : float or numpy.ndarray
        wavelength used for projecting rays to tangent surfaces (from sphere
        model origin).  Defaults to GFA wavelength

    Attributes
    ---------------
    xProj : numpy.ndarray
        x projection of coordinate to the xy plane
    yProj : numpy.ndarray
        y projection of coordinate to xy plane
    distProj : numpy.ndarray
        Distance of projection (mm) proxy for focus offset
        positive when above tangent surface, negative when coord
        is below tangent surface
    obsAngle : float
        Position angle of observation. Angle measured from (image) North
        through East to wok +y. So obsAngle of 45 deg, wok +y points NE.
        Defaults to zero.
    """

    __extra_params__ = ["site", "holeID", "scaleFactor", "obsAngle"]
    __extra_arrays__ = ["wavelength"]
    __computed_arrays__ = ["xProj", "yProj", "distProj"]

    site: Site
    holeID: str
    scaleFactor: float | numpy.ndarray
    obsAngle: float | numpy.ndarray
    wavelength: numpy.ndarray
    xProj: numpy.ndarray
    yProj: numpy.ndarray
    distProj: numpy.ndarray

    def __new__(cls, value, **kwargs):

        verifyWavelength(kwargs, len(value), strict=True)
        verifySite(kwargs, strict=False)

        holeID = kwargs.get("holeID", None)
        if holeID is None:
            raise CoordIOError("Must specify holeID for Tangent Coords")

        scaleFactor = kwargs.get("scaleFactor", None)
        if scaleFactor is None:
            # default to scale factor of 1
            kwargs["scaleFactor"] = 1

        if isinstance(value, Coordinate):
            if "Positioner" in value.coordSysName:
                if holeID not in calibration.VALID_HOLE_IDS:
                    raise CoordIOError("Must be valid holeID for Tangent Coords")
                if holeID.startswith("GFA"):
                    raise CoordIOError("Guide holeID supplied for Positioner coord")
                # going from 2D to 3D coordsys
                # initialize array
                initArr = numpy.zeros((len(value), 3))
                obj = super().__new__(cls, initArr, **kwargs)
                obj._fromPositioner(value)
            elif value.coordSysName == "Guide":
                # going from 2D to 3D coordsys
                # initialize array
                if holeID not in calibration.VALID_HOLE_IDS:
                    raise CoordIOError("Must be valid holeID for Tangent Coords")
                if not holeID.startswith("GFA"):
                    raise CoordIOError(
                        "Cannot convert from guide coords to non-GFA location"
                    )
                initArr = numpy.zeros((len(value), 3))
                obj = super().__new__(cls, initArr, **kwargs)
                obj._fromGuide(value)
            elif value.coordSysName == "Wok":
                # going from 3D to 3D coordsys
                obj = super().__new__(cls, value, **kwargs)
                obj._fromWok(value)
            else:
                raise CoordIOError(
                    "Cannot convert to Tangent from %s" % value.coordSysName
                )

        else:
            obj = super().__new__(cls, value, **kwargs)
            obj._fromRaw()

        return obj

    def _fromPositioner(self, posCoords):
        """Convert from positioner coords to tangent coords

        Parameters
        ------------
        posCoords : `.Positioner`

        """
        if not numpy.isfinite(numpy.sum(posCoords)):
            warnings.warn("NaN values propigated from positioner coordinates",
                          CoordIOUserWarning)
        xTangent, yTangent = conv.positionerToTangent(
            posCoords[:, 0],
            posCoords[:, 1],
            posCoords.xFiber,
            posCoords.yFiber,
            posCoords.alphaArmLength,
            posCoords.alphaOffset,
            posCoords.betaOffset
        )
        self[:, 0] = xTangent
        self[:, 1] = yTangent
        self[:, 2] = 0

        self._fromRaw()

    def _fromGuide(self, guideCoords):
        """Convert from guide coords to tangent coords

        Parameters
        ------------
        guideCoords : `.Guide`
        """
        # print("from guide", self.wavelength)
        if self.holeID not in calibration.VALID_GUIDE_IDS:
            raise CoordIOError(
                "Cannot convert from Guide to (non-guide) wok hole %s" % self.holeID
            )

        # make sure the guide wavelength is specified for all coords
        # this may be a little too extreme of a check
        if numpy.sum(self.wavelength - defaults.INST_TO_WAVE["GFA"]) != 0:
            raise CoordIOError(
                "Cannont convert from Guide coords with non-guide wavelength"
            )

        # note, moved this calculation into
        # conv.guideToTangent, may want to just use that one
        xPix = guideCoords[:, 0]
        yPix = guideCoords[:, 1]
        xBin = guideCoords.xBin
        yBin = guideCoords.yBin

        xTangent = (
            (xPix * xBin - defaults.GFA_CHIP_CENTER)
            * defaults.GFA_PIXEL_SIZE
            / defaults.MICRONS_PER_MM
        )

        yTangent = (
            (yPix * yBin - defaults.GFA_CHIP_CENTER)
            * defaults.GFA_PIXEL_SIZE
            / defaults.MICRONS_PER_MM
        )

        self[:, 0] = xTangent
        self[:, 1] = yTangent
        self[:, 2] = 0  # by definition pixels are in the z=0 tangent plane

        # don't really have to call from raw
        # could just populate proj arrays directly...
        self._fromRaw()

    def _fromWok(self, wokCoords, holeID=None):
        """Convert from wok coords to tangent coords

        Parameters
        -----------
        wokCoords : `.Wok`
        """

        # Converting from wok coordinates we allow holeID to be an array.
        holeID = holeID or self.holeID
        if holeID is None:
            raise CoordIOError("Must be valid holeID for Tangent Coords")

        xWok = wokCoords[:, 0]
        yWok = wokCoords[:, 1]
        zWok = wokCoords[:, 2]

        b, iHat, jHat, kHat = defaults.getHoleOrient(self.site.name, holeID)
        positioner_data = defaults.getPositionerData(self.site.name, holeID)

        tx, ty, tz = conv.wokToTangent(
            xWok, yWok, zWok, b, iHat, jHat, kHat,
            scaleFac=self.scaleFactor,
            dx=positioner_data[9],
            dy=positioner_data[10]
        )

        self[:, 0] = tx
        self[:, 1] = ty
        self[:, 2] = tz

        self._fromRaw()

    def _fromRaw(self):
        """Compute projections to xy plane"""

        rayCenters = _getRayOrigins(
            self.site, self.holeID, self.scaleFactor, self.obsAngle
        )

        for cen, waveCat in zip(rayCenters, ["Apogee", "Boss", "GFA"]):
            arg = numpy.argwhere(self.wavelength == defaults.INST_TO_WAVE[waveCat])

            if len(arg) == 0:
                continue

            arg = arg.squeeze()

            _x, _y, _z = self[arg, 0], self[arg, 1], self[arg, 2]
            xProj, yProj, zProj, distProj = conv.proj2XYplane(_x, _y, _z, cen)
            # note zProj is always zero!
            self.xProj[arg] = xProj
            self.yProj[arg] = yProj
            self.distProj[arg] = distProj


class TangentNoProj(Tangent):
    """Class that doesn't compute projections, intended for internal
    use only to eliminate a recursion problem.

    """

    def _fromRaw(self):
        pass
